// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "DSS3M.h"
#include "MainFrm.h"
#include "DSS3MView.h"

#include "DlgSecond.h"

#include "Global_err.h"
#include "DirectSound.inl"
#include "Mixer.inl"
#include "SoundTrack.inl"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//

LPDIRECTSOUNDBUFFER      lpdsbPrimary;
CSoundTrack* lpS3M     = NULL;
CMixer     * lpMixer   = NULL;
CDSS3MView * lpView    = NULL;

bool         boInTimer = false;
DWORD        dwTicks   = 0;
DWORD        dwBuffers,dwCurBuff;
DWORD        dwBytes;

extern CRITICAL_SECTION CriticalSection;

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNAMIC(CMainFrame, CMDIFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CMDIFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_COMMAND(ID_NEXTORDER, OnNextorder)
	ON_UPDATE_COMMAND_UI(ID_NEXTORDER, OnUpdateNextorder)
	ON_COMMAND(ID_CLEARBUFFER, OnClearbuffer)
	ON_COMMAND(ID_FORCESECOND, OnForcesecond)
	ON_UPDATE_COMMAND_UI(ID_FORCESECOND, OnUpdateForcesecond)
	ON_COMMAND(ID_SECONDSIZE, OnSecondsize)
	ON_COMMAND(ID_TESTSOUND, OnTestsound)
	ON_WM_ACTIVATEAPP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
  m_mmTimer  = 0;
  m_dwPeriod = 0;
  m_lpDSound = NULL;
  m_boForceSecond = false;
}

CMainFrame::~CMainFrame()
{
  hrStop();
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CMDIFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	if (!m_wndToolBar.Create(this) ||
		!m_wndToolBar.LoadToolBar(IDR_MAINFRAME))
	{
		TRACE0("Failed to create toolbar\n");
		return -1;      // fail to create
	}

	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

	// TODO: Remove this if you don't want tool tips or a resizeable toolbar
	m_wndToolBar.SetBarStyle(m_wndToolBar.GetBarStyle() |
		CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC);

	// TODO: Delete these three lines if you don't want the toolbar to
	//  be dockable
	m_wndToolBar.EnableDocking(CBRS_ALIGN_ANY);
	EnableDocking(CBRS_ALIGN_ANY);
	DockControlBar(&m_wndToolBar);

  m_DropTarget.Register(this);                              // drag & drop
	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CMDIFrameWnd::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CMDIFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CMDIFrameWnd::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMainFrame message handlers

HRESULT CMainFrame::hrStop()
  {
  if(m_lpDSound) m_lpDSound->hrStop();
  EnterCriticalSection(&CriticalSection);

  // release timer

  if(m_mmTimer)  timeKillEvent(m_mmTimer);
  if(m_dwPeriod) timeEndPeriod(m_dwPeriod);
  m_mmTimer  = 0;
  m_dwPeriod = 0;

  // release S3M

  lpS3M   = NULL;
  lpMixer = NULL;
  lpView  = NULL;

  // NB: remettre cooperative level ?

  delete m_lpDSound;
  m_lpDSound   = NULL;
  lpdsbPrimary = NULL;

  //

  LeaveCriticalSection(&CriticalSection);

#ifdef _DEBUG
//  AfxMessageBox("DSound stopped");
#endif
  char szBuf[256];
  sprintf(szBuf,"max: %d \ncur: %d",CSoundTrack::dwGetMaxNbParams(),CSoundTrack::dwGetNbParams());
  if(CSoundTrack::dwGetMaxNbParams() > 1) AfxMessageBox(szBuf);
  return(_ERR_(_OK_));
  }

//
//

HRESULT CMainFrame::hrPlay(const DWORD dwSecondKo)
  {
  if(m_lpDSound) return(_ERR_(_OK_));

  // init DSound

  DWORD   dwBuffSize;
  bool    boHard;
  HRESULT hrErr = 0;
  m_lpDSound    = new CDirectSound;
  m_lpDSound->vSetSecondaryKo(dwSecondKo);

  hrErr |= m_lpDSound->hrCreateObject     ();
  hrErr |= m_lpDSound->hrIsPrimaryHardware(&boHard);
  if(m_boForceSecond) 
           m_lpDSound->vSetSecondaryFlag  (true);
           m_lpDSound->vChooseFormat      (44100,true,true);
  hrErr |= m_lpDSound->hrSetBestCoopLevel (AfxGetMainWnd()->m_hWnd);
  hrErr |= m_lpDSound->hrCreatePrimary    ();
  hrErr |= m_lpDSound->hrGetSize          (&dwBuffSize);
  hrErr |= m_lpDSound->hrSetFormat        ();
  hrErr |= m_lpDSound->hrClearBuffer      ();

  if(hrErr)
    {
    hrStop();
    return(hrErr);
    }

  dwBuffers = dwBuffSize / 1024;
  dwCurBuff = 0;
  dwBytes   = m_lpDSound->dwGetBytesPerSample();

  // create timer

  EnterCriticalSection(&CriticalSection);
  lpdsbPrimary = m_lpDSound->lpdsbGetBuffer();
  LeaveCriticalSection(&CriticalSection);

  m_dwPeriod = 1024*1000/m_lpDSound->dwGetBytesPerSec();    // 1Ko segment each time
  if(timeBeginPeriod(m_dwPeriod) != TIMERR_NOERROR)
    {
    hrStop();
    return(-1);
    }

  dwTicks  = 0;
  m_mmTimer  = timeSetEvent(m_dwPeriod,1,TimerFunc,0,TIME_PERIODIC);
  if(!m_mmTimer)
    {
    hrStop();
    return(-1);
    }

  // play

  if(hrErr = m_lpDSound->hrPlay())
    {
    hrStop();
    return(hrErr);
    }

#ifdef _DEBUG
//  AfxMessageBox("DSound playing");
#endif
  return(_ERR_(_OK_));
  }

//

HRESULT CMainFrame::hrPlayDSound()
  {
  if(!m_lpDSound) return -1;
  return m_lpDSound->hrPlay();
  }

//

HRESULT CMainFrame::hrStopDSound()
  {
  if(!m_lpDSound) return -1;
  HRESULT hrErr = m_lpDSound->hrStop();
  m_lpDSound->hrClearBuffer();
  return hrErr;
  }

//
// timer callback
//

void CALLBACK TimerFunc(UINT uTimerID,UINT uMsg,DWORD dwUser,DWORD dw1,DWORD dw2)
  {
  if(boInTimer) return;                                     // avoid re-entry

  EnterCriticalSection(&CriticalSection);

  if(!lpdsbPrimary)              { LeaveCriticalSection(&CriticalSection); return; }
  if(!lpS3M)                     { LeaveCriticalSection(&CriticalSection); return; }
  if(!lpMixer)                   { LeaveCriticalSection(&CriticalSection); return; }
  if(!lpView)                    { LeaveCriticalSection(&CriticalSection); return; }
  if(!CMixer::lpbGetDMABuffer()) { LeaveCriticalSection(&CriticalSection); return; }

  boInTimer = true;

  //

  dwTicks++;

  DWORD   dwPlay,dwWrite,dwKo;
  DWORD   dwStatus;
  HRESULT hr = lpdsbPrimary->GetStatus(&dwStatus);
  if(dwStatus & DSBSTATUS_BUFFERLOST)
    {
    hr = lpdsbPrimary->Restore();
    if(FAILED(hr))
      {
      boInTimer = false;
      LeaveCriticalSection(&CriticalSection);
      return;
      }
    hr = lpdsbPrimary->Play(0,0,DSBPLAY_LOOPING);
    }

  hr   = lpdsbPrimary->GetCurrentPosition(&dwPlay,&dwWrite);
  if(FAILED(hr))
    {
    boInTimer = false;
    LeaveCriticalSection(&CriticalSection);
    return;
    }
  dwKo = dwWrite / 1024;

  if((dwCurBuff+dwBuffers-dwKo) % dwBuffers <= dwBuffers/2) do
    {
    static BYTE bRow = 0;

    CMixer::vClearDMABuffer();
    lpMixer->hrHandler();

    if((bRow != lpS3M->bGetRow()) && lpView) lpView->PostMessage(WM_UPDATE_CTRLS,0,0);

    // copy

    LPVOID lpvPtr1, lpvPtr2;
    DWORD  dwBytes1,dwBytes2;

    hr = lpdsbPrimary->Lock(dwCurBuff*1024,1024*dwBytes,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);

    if(hr == DSERR_BUFFERLOST)
      {
      hr = lpdsbPrimary->Restore();
      hr = lpdsbPrimary->Lock(dwCurBuff*1024,1024*dwBytes,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);
      }
    if(FAILED(hr));

    else
      {
      DWORD dw1 = 1024*dwBytes;
      DWORD dw2 = 0;
      if(dw1 > dwBytes1)
        {
        dw2 = dw1-dwBytes1;
        dw1 = dwBytes1;
        }

      memcpy(lpvPtr1,CMixer::lpbGetDMABuffer(),    dw1);
      memcpy(lpvPtr2,CMixer::lpbGetDMABuffer()+dw1,dw2);

      hr = lpdsbPrimary->Unlock(lpvPtr1,dwBytes1,lpvPtr2,dwBytes2);
      }

    //

    dwCurBuff += dwBytes;
    dwCurBuff %= dwBuffers;
    }
    while((dwCurBuff+dwBuffers-dwKo) % dwBuffers <= dwBuffers/2);

  //

  LeaveCriticalSection(&CriticalSection);
  boInTimer = false;
  }

//
//

void CMainFrame::OnNextorder() 
  {
  if(!lpS3M->boIsPlaying()) return;
  EnterCriticalSection(&CriticalSection);
  lpS3M->hrNextOrder();
  LeaveCriticalSection(&CriticalSection);
  }

void CMainFrame::OnUpdateNextorder(CCmdUI* pCmdUI) 
  {
  EnterCriticalSection(&CriticalSection);
  pCmdUI->Enable(lpS3M && lpS3M->boIsPlaying());
  LeaveCriticalSection(&CriticalSection);
  }

//
//

void CMainFrame::OnClearbuffer() 
  {
  if(m_lpDSound) m_lpDSound->hrClearBuffer();
  }

//
//

void CMainFrame::OnForcesecond() 
  {
  m_boForceSecond = !m_boForceSecond;
  hrReinit();
  }

//

HRESULT CMainFrame::hrReinit()
  {
  EnterCriticalSection(&CriticalSection);

  CSoundTrack* pS3M   = lpS3M;
  CMixer     * pMixer = lpMixer;
  CDSS3MView * pView  = lpView;
  DWORD        dwKo   = m_lpDSound->dwGetSecondaryKo();

  hrStop();
  HRESULT hrErr = hrPlay(dwKo);

  lpS3M   = pS3M;
  lpMixer = pMixer;
  lpView  = pView;

  LeaveCriticalSection(&CriticalSection);
  return(hrErr);
  }

//

void CMainFrame::OnUpdateForcesecond(CCmdUI* pCmdUI) 
  {
  bool boSecond = m_boForceSecond;
  if(m_lpDSound)  boSecond |= m_lpDSound->boGetSecondaryFlag();

  pCmdUI->SetCheck(boSecond);
  }

//

void CMainFrame::OnSecondsize() 
  {
  if(!m_lpDSound) return;
  CDlgSecond dlgSecond;
  dlgSecond.m_dwSizeKo = m_lpDSound->dwGetSecondaryKo();
  if(dlgSecond.DoModal() != IDOK) return;

  if(!m_lpDSound) return;
  m_lpDSound->vSetSecondaryKo(dlgSecond.m_dwSizeKo);

  if(m_lpDSound->boGetSecondaryFlag()) hrReinit();
  }

//

void CMainFrame::OnTestsound() 
  {
  static BYTE bOrdRow[13][2] = { {37,0}, {37,32}, {37,48}, {37,56},
                                 {38,0}, {38,4},  {38,12}, {38,20}, {38,32}, {38,50},
                                 {39,0}, {39,8},  {39,20} };
  int iRand = rand() % 13;

  EnterCriticalSection(&CriticalSection);
  if(m_lpDSound && lpMixer)
    lpMixer->hrStartPlaying(bOrdRow[iRand][0],bOrdRow[iRand][1],m_lpDSound->dwGetFreq(),m_lpDSound->boGetStereo(),m_lpDSound->boGet16Bits(),false,true);
  LeaveCriticalSection(&CriticalSection);
	}

//

void CMainFrame::OnActivateApp(BOOL bActive, HTASK hTask) 
  {
	CMDIFrameWnd::OnActivateApp(bActive, hTask);
	
  if(bActive)
    {
    if(m_lpDSound) HRESULT hrErr = m_lpDSound->hrSetCooperative(AfxGetMainWnd()->m_hWnd,DSSCL_WRITEPRIMARY);
    }
  }
