/**********************************************************************************************************************/
/*                                                ERROR STRING MANAGER                                                */
/* 08/06/00, M : creation of the file                                                                                 */
/**********************************************************************************************************************/

#ifndef   _ERRORSTRING_H_
#define   _ERRORSTRING_H_

/************************************************ DEFINE **************************************************************/

#define   _OK_                          0x0                           // same as HRESULT S_OK
#define   _HR_FAILED_                   0x80000000

#define   _GLOBAL_                     (0x00000000+_HR_FAILED_)
#define   _MYTHOS_                     (0x01000000+_HR_FAILED_)       // coder=0x01

// release => do nothing

#ifndef   _DEBUG

#define   _ERR_(hrID)                   hrID                          // simply return ID
#define   _ERROR_STRING_(hrID,szString)                               // no error string

#else

// debug

#pragma   warning(disable:4786)                                       // otherwise STL lines are too long
#include  <map>
typedef   unsigned long                 DWORD;

/************************************************ CLASSES *************************************************************/

/**********************************************************************************************************************/
/*                                                CErrors                                                             */
/* this class stores the ID->string map                                                                               */
/**********************************************************************************************************************/

class     CErrors
  {
  typedef std::map< DWORD,char*,std::less<DWORD> >          mapDW2string;
  typedef mapDW2string::iterator                            mapDW2stringIterator;
  typedef std::pair<mapDW2stringIterator,bool>              pairDW2string;

// public methods

  public:
                   ~CErrors             (void);
    bool            boAddError          (const DWORD dwID,const char* lpszString);
    const char*     lpszGetErrorString  (const DWORD dwID);

// public members

  public:

    static DWORD    m_dwNbErrors;                           // nb entries in map
    static bool     m_boDuplicatedID;                       // duplicated string ID ?
    static DWORD    m_dwLastDuplicated;                     // last duplicated ID stored

// protected members

  protected:

    mapDW2string    m_mapErrors;                            // ID->string map
  };

/**********************************************************************************************************************/

//
// destructor: delete the strings & the map
//

inline CErrors::~CErrors()
  {
  for(mapDW2stringIterator iter = m_mapErrors.begin(); iter != m_mapErrors.end(); iter++)
    delete [] (*iter).second;
  m_mapErrors.clear();
  }

//
// boAddError: add a string to the map, check for duplicated ID
//

inline bool CErrors::boAddError(const DWORD dwID,const char* lpszString)
  {
  char*  lpszStr = new char[strlen(lpszString)+1];
  strcpy(lpszStr,lpszString);

  pairDW2string pairResult = m_mapErrors.insert(std::make_pair(dwID,lpszStr));
  if(!pairResult.second) 
    {
    delete [] lpszStr;
    m_boDuplicatedID   = true;
    m_dwLastDuplicated = dwID;
    }
  else m_dwNbErrors++;

  return pairResult.second;
  }

//
// lpszGetErrorString: return string corresponding to given ID
//

inline const char* CErrors::lpszGetErrorString(const DWORD dwID)
  {
  mapDW2stringIterator iter = m_mapErrors.find(dwID);
  if(iter == m_mapErrors.end()) return(NULL);
  return (*iter).second;
  }

/**********************************************************************************************************************/
/*                                                CError                                                              */
/* this class creates the CErrors object, and adds the strings                                                        */
/**********************************************************************************************************************/

class     CError
  {
// public methods

  public:

                    CError              (const DWORD dwID,const char* lpszString);
                   ~CError              (void);
    static DWORD    dwGetError          (const DWORD dwID);

// public members

  public:

    static DWORD                        m_dwLastError;      // last error passed through _ERR_ (put it in the watch window!)
    static const char*                  m_lpszLastError;    // last error string               (put it in the watch window!)

// protected members

  protected:

    static CErrors*                     m_lpErrors;         // CErrors (map) object
  };

/**********************************************************************************************************************/

//
// constructor: creates m_lpErrors the 1st time, adds a new string to the map
//

inline CError::CError(const DWORD dwID,const char* lpszString)
  {
  if(!m_lpErrors) m_lpErrors = new CErrors;
  if( m_lpErrors) m_lpErrors->boAddError(dwID,lpszString);
  }

//
// destructor: delete the map object
//

inline CError::~CError()
  {
  if(m_lpErrors) delete m_lpErrors;
  m_lpErrors = NULL;
  }

//
// dwGetError: called by _ERR_, return the ID but look for the corresponding string before
//

inline DWORD CError::dwGetError(const DWORD dwID)
  {
  if(!(dwID & _HR_FAILED_)) return dwID;                    // not an error
  m_dwLastError = dwID;
  if(m_lpErrors) m_lpszLastError = CError::m_lpErrors->lpszGetErrorString(dwID);
  return dwID;
  }

/************************************************ MACROS **************************************************************/

// _ERR_: used to find the error string, this way:
//        HRESULT hrErr = _ERR_(_MY_ERROR_ID_);
//        return _ERR_(_MY_ERROR_ID_);

#define   _ERR_(hrID)                   CError::dwGetError(hrID)

// _ERROR_STRING_: used to define the strings in the cpp (ex: RendererGL_err.cpp), like this:
//                 #include "RendererGL_err.h"
//                 _ERROR_STRING_(_GLERR_MYERRORID0_,"GL error number 0")

#define   _ERROR_STRING_(hrID,szString) CError error##hrID(hrID,szString);

/**********************************************************************************************************************/

#endif    // _DEBUG
#endif    // _ERRORSTRING_H_
