/**********************************************************************************************************************/
/*                                                DIRECTSOUND CLASS                                                   */
/* 03/10/99, M : creation of the file                                                                                 */
/*               provides access to DirectSound primary or secondary buffer to play                                   */
/* 28/10/99, M : class reorg                                                                                          */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "DirectSound.h"
#include  "DirectSound_err.h"

#include  "Debug.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// 03/10/99, M
/**********************************************************************************************************************/

CDirectSound::CDirectSound()
  {
  _DTRCDEF_("CDirectSound::CDirectSound");
  vInit();
  }

/************************************************ destructor **********************************************************/
// 03/10/99, M
/**********************************************************************************************************************/

CDirectSound::~CDirectSound()
  {
  _DTRCDEF_("CDirectSound::~CDirectSound");
  vClean();
  }

/************************************************ vInit ***************************************************************/
// initialize members & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CDirectSound::vInit()
  {
  _DTRCDEF_("CDirectSound::vInit");
  m_lpdsDSound     = NULL;
  m_lpdsbPrimary   = NULL;
  m_boSecondary    = false;
  m_dwSecondKo     = 64;

  m_boStereo = true;
  m_bo16Bits = true;
  m_dwFreq   = 44100;
  }

/************************************************ vClean **************************************************************/
// release memory & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CDirectSound::vClean()
  {
  _DTRCDEF_("CDirectSound::vClean");
  if(m_lpdsDSound) m_lpdsDSound->Release();                 // automatically releases any existing buffers
  vInit();
  }

/**********************************************************************************************************************/
/*                                                INITS                                                               */
/**********************************************************************************************************************/

/************************************************ hrCreateObject ******************************************************/
// create DSound object
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrCreateObject()
  {
  _DTRCDEF_("CDirectSound::hrCreateObject");
  HRESULT hrErr = DirectSoundCreate(NULL,&m_lpdsDSound,NULL);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_NODSOUND_));
  else              return(_ERR_(_OK_));
  }

/************************************************ hrSetCooperative ****************************************************/
// try to change cooperative level
// 14/11/99, M
// in : main window handle,wanted level
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrSetCooperative(const HWND hWnd,const DWORD dwLevel)
  {
  _DTRCDEF_("CDirectSound::hrSetCooperative");
  if(!m_lpdsDSound) return(_ERR_(_S3MERR_NODSOUND_));

  HRESULT hrErr = m_lpdsDSound->SetCooperativeLevel(hWnd,dwLevel);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_COOPLEVEL_));
  else              return(_ERR_(_OK_));
  }

/************************************************ hrSetBestCoopLevel **************************************************/
// try WRITEPRIMARY first, else EXCLUSIVE
// 14/11/99, M
// in : main window handle
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrSetBestCoopLevel(const HWND hWnd)
  {
  _DTRCDEF_("CDirectSound::hrSetBestCoopLevel");
  if(!m_boSecondary && !hrSetCooperative(hWnd,DSSCL_WRITEPRIMARY)) return(_ERR_(_OK_));

  m_boSecondary = true;
  return(hrSetCooperative(hWnd,DSSCL_EXCLUSIVE));
  }

/************************************************ hrIsPrimaryHardware *************************************************/
// check primary buffer hardware support
// 14/11/99, M
// in : where to store the result
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrIsPrimaryHardware(bool* lpboResult)
  {
  _DTRCDEF_("CDirectSound::hrIsPrimaryHardware");
  if(!m_lpdsDSound)  return(_ERR_(_S3MERR_NODSOUND_));

  DSCAPS dscaps;
  ZeroMemory(&dscaps,sizeof(DSCAPS));
  dscaps.dwSize    = sizeof(DSCAPS);

  HRESULT hrErr = m_lpdsDSound->GetCaps(&dscaps);
  if(FAILED(hrErr))  return(_ERR_(_S3MERR_DSCAPS_));

  if(dscaps.dwFlags & DSCAPS_EMULDRIVER) *lpboResult = false;
  else                                   *lpboResult = true;
  m_boSecondary = !*lpboResult;
  return(_ERR_(_OK_));
  }

/************************************************ vChooseFormat *******************************************************/
// choose buffer format parameters
// 14/11/99, M
// in : frequency,16b & stereo flags
// out:
/**********************************************************************************************************************/

void CDirectSound::vChooseFormat(const DWORD dwFreq,const bool bo16Bits,const bool boStereo)
  {
  _DTRCDEF_("CDirectSound::vChooseFormat");
  m_dwFreq   = dwFreq;
  m_bo16Bits = bo16Bits;
  m_boStereo = boStereo;
  }

/************************************************ hrCreatePrimary *****************************************************/
// create primary buffer (or secondary if emulated)
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrCreatePrimary()
  {
  _DTRCDEF_("CDirectSound::hrCreatePrimary");
  if(!m_lpdsDSound)   return(_ERR_(_S3MERR_NODSOUND_));

  DSBUFFERDESC dsbdesc;
  ZeroMemory(&dsbdesc,sizeof(DSBUFFERDESC));
  dsbdesc.dwSize    = sizeof(DSBUFFERDESC);
  dsbdesc.dwFlags   = DSBCAPS_PRIMARYBUFFER | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_STICKYFOCUS;

  WAVEFORMATEX wfx;

  if(m_boSecondary)
    {
    memset(&wfx,0,sizeof(WAVEFORMATEX));
    wfx.wFormatTag      = WAVE_FORMAT_PCM;
    wfx.nChannels       = WORD(m_boStereo ?  2 : 1);          // mono/stro
    wfx.nSamplesPerSec  = m_dwFreq;                           // frequency
    wfx.wBitsPerSample  = WORD(m_bo16Bits ? 16 : 8);          // 8b/16b
    wfx.nBlockAlign     = WORD(wfx.wBitsPerSample/8 * wfx.nChannels);
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

    dsbdesc.dwFlags       = DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | 
                            DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_STICKYFOCUS;
    dsbdesc.dwBufferBytes = m_dwSecondKo*1024;
    dsbdesc.lpwfxFormat   = &wfx;
    }

  HRESULT hrErr = m_lpdsDSound->CreateSoundBuffer(&dsbdesc,&m_lpdsbPrimary,NULL);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_PRIMARY_));
  else              return(_ERR_(_OK_));
  }

/************************************************ hrSetFormat *********************************************************/
// set buffer format
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrSetFormat()
  {
  _DTRCDEF_("CDirectSound::hrSetFormat");
  if(!m_lpdsbPrimary) return(_ERR_(_S3MERR_NOPRIMARY_));

  WAVEFORMATEX wfx;
  memset(&wfx,0,sizeof(WAVEFORMATEX));
  wfx.wFormatTag      = WAVE_FORMAT_PCM;
  wfx.nChannels       = WORD(m_boStereo ?  2 : 1);          // mono/stro
  wfx.nSamplesPerSec  = m_dwFreq;                           // frequency
  wfx.wBitsPerSample  = WORD(m_bo16Bits ? 16 : 8);          // 8b/16b
  wfx.nBlockAlign     = WORD(wfx.wBitsPerSample/8 * wfx.nChannels);
  wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
  m_dwBytesPerSample  = wfx.nBlockAlign;

  if(m_boSecondary) return(_ERR_(_OK_));
  HRESULT hrErr = m_lpdsbPrimary->SetFormat(&wfx);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_SETFORMAT_));
  else              return(_ERR_(_OK_));
  }

/************************************************ hrGetSize ***********************************************************/
// get buffer size
// 14/11/99, M
// in : where to store the size
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrGetSize(DWORD* lpdwSize)
  {
  _DTRCDEF_("CDirectSound::hrGetSize");
  if(!m_lpdsbPrimary) return(_ERR_(_S3MERR_NOPRIMARY_));

  DSBCAPS dsbcaps;
  ZeroMemory(&dsbcaps,sizeof(DSBCAPS));
  dsbcaps.dwSize    = sizeof(DSBCAPS);

  HRESULT hrErr = m_lpdsbPrimary->GetCaps(&dsbcaps);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_SETFORMAT_));

  *lpdwSize = m_dwBufferSize = dsbcaps.dwBufferBytes;
  return(_ERR_(_OK_));
  }

/************************************************ hrClearBuffer *******************************************************/
// initialize buffer to silence
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrClearBuffer()
  {
  _DTRCDEF_("CDirectSound::hrClearBuffer");
  if(!m_lpdsbPrimary) return(_ERR_(_S3MERR_NOPRIMARY_));

  LPVOID  lpvPtr1, lpvPtr2;
  DWORD   dwBytes1,dwBytes2;
  HRESULT hrErr = m_lpdsbPrimary->Lock(0,m_dwBufferSize,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);

  if(hrErr == DSERR_BUFFERLOST)
    {
    m_lpdsbPrimary->Restore();
    hrErr = m_lpdsbPrimary->Lock(0,m_dwBufferSize,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);
    }
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_CLEARBUFFER_));

  memset(lpvPtr1,m_bo16Bits ? 0 : 128,dwBytes1);

  hrErr = m_lpdsbPrimary->Unlock(lpvPtr1,dwBytes1,lpvPtr2,dwBytes2);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_CLEARBUFFER_));
  else              return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PLAY                                                                */
/**********************************************************************************************************************/

/************************************************ hrPlay **************************************************************/
// start playing buffer
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrPlay()
  {
  _DTRCDEF_("CDirectSound::hrPlay");
  if(!m_lpdsbPrimary) return(_ERR_(_S3MERR_NOPRIMARY_));
  
  HRESULT hrErr = m_lpdsbPrimary->Play(0,0,DSBPLAY_LOOPING);
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_DSPLAY_));
  else              return(_ERR_(_OK_));
  }

/************************************************ hrStop **************************************************************/
// stop playing buffer
// 14/11/99, M
// in :
// out: 0=OK or error ID
/**********************************************************************************************************************/

HRESULT CDirectSound::hrStop()
  {
  _DTRCDEF_("CDirectSound::hrStop");
  if(!m_lpdsbPrimary) return(_ERR_(_S3MERR_NOPRIMARY_));
  
  HRESULT hrErr = m_lpdsbPrimary->Stop();
  if(FAILED(hrErr)) return(_ERR_(_S3MERR_DSSTOP_));
  else              return(_ERR_(_OK_));
  }
