/**********************************************************************************************************************/
/*                                                IT CLASS                                                            */
/* 27/12/99, M : creation of the file                                                                                 */
/*               handles an IT soundtrack                                                                             */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "IT.h"
#include  "Global_err.h"
#include  "SoundTrack_err.h"

#include  "Debug.h"
#include  "Instrument.h"
#include  "ITSample.h"
#include  "Sample.inl"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/************************************************ STATIC MEMBERS ******************************************************/

const char CIT::m_szExtension[_MAX_EXT] = ".IT";            // IT file extension

static BYTE bOldMask [_MAX_CHANNELS_];
static BYTE bOldNote [_MAX_CHANNELS_];
static BYTE bOldInstr[_MAX_CHANNELS_];
static BYTE bOldVol  [_MAX_CHANNELS_];
static BYTE bOldPan  [_MAX_CHANNELS_];
static BYTE bOldComm [_MAX_CHANNELS_];
static BYTE bOldInfo [_MAX_CHANNELS_];

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// 27/12/99, M
/**********************************************************************************************************************/

CIT::CIT()
  {
  _DTRCDEF_("CIT::CIT");
  //vInit();
  }

/************************************************ destructor **********************************************************/
// 27/12/99, M
/**********************************************************************************************************************/

CIT::~CIT()
  {
  _DTRCDEF_("CIT::~CIT");
  //vClean();
  }

/************************************************ vInit ***************************************************************/
// initialize members & resources
// 27/12/99, M
// in :
// out:
/**********************************************************************************************************************/

void CIT::vInit()
  {
  _DTRCDEF_("CIT::vInit");
  CSoundTrack::vInit();
  }

/************************************************ vClean **************************************************************/
// release memory & resources
// 27/12/99, M
// in :
// out:
/**********************************************************************************************************************/

void CIT::vClean()
  {
  _DTRCDEF_("CIT::vClean");
  CSoundTrack::vClean();
  }

/**********************************************************************************************************************/
/*                                                LOAD & INITS                                                        */
/**********************************************************************************************************************/

/************************************************ lpszGetExtension ****************************************************/
// return IT extension
// 27/12/99, M
// in :
// out: IT extension
/**********************************************************************************************************************/

const char* CIT::lpszGetExtension()
  {
  _DTRCDEF_("CIT::lpszGetExtension");
  return(m_szExtension);
  }

/************************************************ hrCheckSignature ****************************************************/
// check the current file is an IT
// 27/12/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CIT::hrCheckSignature()
  {
  _DTRCDEF_("CIT::hrCheckSignature");
  HRESULT hrError = CSoundTrack::hrCheckSignature();
  if(hrError) return(hrError);

  // compare IT signature

  IT_HEADER* lpHeader = (IT_HEADER*)m_lpbFile;
  if(strncmp(lpHeader->szIMPM,"IMPM",4) != 0)
    {
    _DWAR_("ERROR: not an IT file");
    return(m_hrError = _ERR_(_S3MERR_BADFORMAT_));
    }

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

/************************************************ hrInitVars **********************************************************/
// init starting values + samples & patterns pointers
// 27/12/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CIT::hrInitVars()
  {
  _DTRCDEF_("CIT::hrInitVars");
  HRESULT hrError = CSoundTrack::hrInitVars();
  if(hrError) return(hrError);

  // starting values

  IT_HEADER* lpHeader = (IT_HEADER*)m_lpbFile;

  m_bPatternSpeed0 = lpHeader->bInitSpeed;
  m_bTempo0        = lpHeader->bInitTempo;
  m_bGlobalVolume0 = BYTE(lpHeader->bGlobalVol >> 1);
  m_bMasterVolume  = lpHeader->bMasterVol;
  m_boStereo0      =(lpHeader->wFlags & _ITFLAG_STEREO_) != 0;

  // balance

  m_dwChannels  = 0;

  for(DWORD dwI = 0; dwI < 64; dwI++)
    {                                                       // 64 channels
    BYTE bChanVal       = lpHeader->bChanPan[dwI];
    m_boChanActive[dwI] = (bChanVal < _ITFLAG_CHANDISABLED_);
    if(m_boChanActive[dwI]) 
      {
      m_dwChannels = dwI+1;

      if(m_boStereo0) m_bPanPos0[dwI] = BYTE(bChanVal/4.2); // [0;15]
      else            m_bPanPos0[dwI] = 8;
      }
    }

  // orders

  BYTE* lpbOrders = m_lpbFile + 0xc0;

  for(dwI = 0; dwI < _MAX_ORDERS_;      dwI++) m_bOrderList[dwI] = 255;
  for(dwI = 0; dwI < lpHeader->wOrdNum; dwI++) m_bOrderList[dwI] = *(lpbOrders+dwI);

  m_bOrders = BYTE(lpHeader->wOrdNum);

  // samples

  DWORD* lpdwSmp = (DWORD*)(lpbOrders + lpHeader->wOrdNum + lpHeader->wInstNum*4);

  for(dwI = 0; dwI < _MAX_INSTRUMENTS_; dwI++) m_lpInstrument[dwI] = NULL;
  for(dwI = 0; dwI < lpHeader->wSmpNum; dwI++)
    {
    // header

    DWORD         dwParaPtr   = *(lpdwSmp+dwI);
    ITSMP_HEADER* lpSMPHeader = (ITSMP_HEADER*)(m_lpbFile + dwParaPtr);

    CITSample  *  lpSample    = new CITSample;
    CInstrument*  lpInstr     = new CInstrument(lpSample);  // mono-sample instrument
    if(!lpSample || !lpInstr)
      {
      _DWAR_("ERROR: out of memory");
      m_boInitVars = false;
      return(m_hrError = _ERR_(_GLOBERR_OUTOFMEMORY_));
      }
    m_lpInstrument[dwI] = lpInstr;

    // data

    if(lpSMPHeader->bFlags & _ITSMPFLAG_WAVE_)
      {                                                     // there is a sample
      dwParaPtr = lpSMPHeader->dwOffset;
      lpSample->vSetHeader (lpSMPHeader);                   // header   address
      lpSample->vSetDataRAM(m_lpbFile+dwParaPtr);           // data RAM address
      lpSample->vSetC4Speed(lpSMPHeader->dwC5Speed/2);      // C4 speed
      }
    else
      {                                                     // no sample
      lpSample->vSetHeader (NULL);
      lpSample->vSetDataRAM(NULL);
      lpSample->vSetC4Speed(0);
      }

    // other variables

    if(lpSMPHeader->bFlags & _ITSMPFLAG_LOOP_)
      {                                                     // loop
      lpSample->vSetRepeat(lpSMPHeader->dwLoopStart);
      lpSample->vSetLength(lpSMPHeader->dwLength);
      lpSample->vSetRepLen(lpSMPHeader->dwLoopEnd - lpSMPHeader->dwLoopStart);
      }
    else
      {
      lpSample->vSetLength(lpSMPHeader->dwLength);
      lpSample->vSetRepLen(0);
      }

    lpSample->vSet16Bits  ((lpSMPHeader->bFlags & _ITSMPFLAG_16BITS_)   != 0);
    lpSample->vSetVolume   (lpSMPHeader->bVolume);
    lpSample->vSetPingPong((lpSMPHeader->bFlags & _ITSMPFLAG_PINGPONG_) != 0);

    // unsigned->signed conversion

    if(lpSample->lpbGetDataRAM())
      {
      DWORD dwLen   = lpSample->dwGetLength  ();
      BYTE* lpbData = lpSample->lpbGetDataRAM();
      WORD* lpwData = reinterpret_cast<WORD*>(lpbData);
      DWORD dwSmp;

      if(lpSample->boGet16Bits())
        {
        for(dwSmp = 0; dwSmp < dwLen; dwSmp++) *lpwData++ += 0x8000;
        }
      else
        {
        for(dwSmp = 0; dwSmp < dwLen; dwSmp++) *lpbData++ += 0x80;
        }
      }

    // length variables in bytes

    if(lpSample->boGet16Bits())
      {
      lpSample->vSetRepeat(2*lpSample->dwGetRepeat());
      lpSample->vSetLength(2*lpSample->dwGetLength());
      lpSample->vSetRepLen(2*lpSample->dwGetRepLen());
      }
    }

  // patterns

  DWORD* lpdwPatt = lpdwSmp + lpHeader->wSmpNum;

  for(dwI = 0; dwI < _MAX_PATTERNS_;     dwI++) m_lpbPattern[dwI] = NULL;
  for(dwI = 0; dwI < lpHeader->wPattNum; dwI++)
    {
    DWORD dwParaPtr = *(lpdwPatt+dwI);

    if(dwParaPtr)
      {                                                     // not empty
      WORD wDataLength  = *(m_lpbFile+dwParaPtr);
      WORD wPattRows    = *(m_lpbFile+dwParaPtr+2);
      m_lpbPattern[dwI] = 8+m_lpbFile+dwParaPtr;            // skip pattern header
      }
    }

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PATTERN DATA                                                        */
/**********************************************************************************************************************/

/************************************************ vGoToRow ************************************************************/
// go to given row in packed pattern
// 27/12/99, M
// in : pattern start address,wanted row
// out: pattern row   address in lplpbPattern
/**********************************************************************************************************************/

void CIT::vGoToRow(BYTE** lplpbPattern,const BYTE bRow)
  {
  _DTRCDEF_("CIT::vGoToRow");
  if(!*lplpbPattern) return;

  BYTE  bRowsToGo  = bRow;
  BYTE* lpbPattern = *lplpbPattern;

  while(bRowsToGo)
    {
    BYTE bVal = *lpbPattern++;

    if(bVal)
      {                                                     // 0 = empty row
      BYTE bChan = BYTE((bVal-1) & 63);
      BYTE bMask = bOldMask[bChan];
      if(bVal  & 128) bMask = *lpbPattern++;                // mask
      bOldMask[bChan] = bMask;

      if(bMask &   1) lpbPattern ++;                        // note
      if(bMask &   2) lpbPattern ++;                        // instrument
      if(bMask &   4) lpbPattern ++;                        // volume/panning
      if(bMask &   8) lpbPattern += 2;                      // command & info
      }

    else bRowsToGo--;                                       // next row
    }

  *lplpbPattern = lpbPattern;
  }

/************************************************ vUnpackRow **********************************************************/
// unpack one row
// 27/12/99, M
// in : where to store unpacked row,packed data address (modified)
// out: 
/**********************************************************************************************************************/

void CIT::vUnpackRow(BYTE*  lpbUnpacked,BYTE** lplpbNote)
  {
  _DTRCDEF_("CIT::vUnpackRow");
  if(!*lplpbNote) return;
  BYTE* lpbNote = *lplpbNote;

  while(true)
    {
    BYTE bVal = *lpbNote++;

    if(bVal)
      {                                                     // 0 = empty row
      BYTE  bChannel = BYTE((bVal-1) & 63);
      BYTE* lpbChann = lpbUnpacked + bChannel*5;
      BYTE  bMask    = bOldMask[bChannel];
      if(bVal  & 128)  bMask = *lpbNote++;                  // mask
      bOldMask[bChannel] = bMask;

      if(bMask &   1) 
        {
        BYTE bITNote  = *lpbNote++;
        BYTE bOctave  = BYTE(bITNote / 12);
        BYTE bNote    = BYTE(bITNote % 12);
        BYTE bS3MNote = BYTE((bOctave << 4)+bNote);
        bOldNote [bChannel] = *(lpbChann+0) = bS3MNote;     // note

        if(bITNote == 254)
          bOldNote [bChannel] = *(lpbChann+0) = bITNote;    // ^^
        }

      if(bMask &   2) 
        bOldInstr[bChannel] = *(lpbChann+1) = *lpbNote++;   // instrument

      if(bMask &   4)
        {
        BYTE bVolPan = *lpbNote++;
        if(bVolPan & 128)
          bOldPan[bChannel] = bVolPan;                      // panning
        else 
          bOldVol[bChannel] = *(lpbChann+2) = bVolPan;      // volume
        }

      if(bMask &   8)
        {
        bOldComm[bChannel]  = *(lpbChann+3) = *lpbNote++;   // command
        bOldInfo[bChannel]  = *(lpbChann+4) = *lpbNote++;   // info
        }

      if(bMask &  16) *(lpbChann+0) = bOldNote [bChannel];  // old note
      if(bMask &  32) *(lpbChann+1) = bOldInstr[bChannel];  // old instrument
      if(bMask &  64) *(lpbChann+2) = bOldVol  [bChannel];  // old volume
      if(bMask & 128)
        {
        *(lpbChann+3) = bOldComm[bChannel];                 // old command
        *(lpbChann+4) = bOldInfo[bChannel];                 // old info
        }
      }

    else
      {
      *lplpbNote = lpbNote;
      break;
      }
    }
  }

/************************************************ lpszGetTitle ********************************************************/
// return song title
// 27/12/99, M
// in :
// out: title address
/**********************************************************************************************************************/

const char* CIT::lpszGetTitle() const
  {
  ((IT_HEADER*)m_lpbFile)->szITName[25] = '\0';
  return((const char*)((IT_HEADER*)m_lpbFile)->szITName);
  }
