/**********************************************************************************************************************/
/*                                                IT CLASS                                                            */
/* 27/12/99, M : creation of the file                                                                                 */
/*               handles an IT soundtrack                                                                             */
/**********************************************************************************************************************/

#if !defined(AFX_IT_H__35F76920_BC62_11D3_9848_D0068C1C3643__INCLUDED_)
#define AFX_IT_H__35F76920_BC62_11D3_9848_D0068C1C3643__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif

/************************************************ INCLUDE *************************************************************/

#include  "SoundTrack.h"                                    // base class

/************************************************ DEFINE **************************************************************/

// header flags

#define   _ITFLAG_STEREO_               1                   // stereo ON
#define   _ITFLAG_VOL0MIXOPTIM_         2
#define   _ITFLAG_INSTR_                4                   // use instruments or samples
#define   _ITFLAG_LINEARSLIDE_          8
#define   _ITFLAG_OLDEFFECTS_           16                  // old vs IT effects
#define   _ITFLAG_CHANDISABLED_         128                 // muted channel

/************************************************ TYPEDEF & STRUCT ****************************************************/

// IT file header

#pragma   pack(1)

typedef   struct    {
          char      szIMPM   [ 4];                          // "IMPM" signature
          char      szITName [26];                          // song name (ASCIIZ)
          char      szUnused [ 2];
          WORD      wOrdNum;                                // number of orders
          WORD      wInstNum;                               // number of instruments
          WORD      wSmpNum;                                // number of samples
          WORD      wPattNum;                               // number of patterns
          WORD      wTracker;                               // created    with tracker
          WORD      wCompatible;                            // compatible with tracker
          WORD      wFlags;                                 // +1=stereo; +4=instruments; +16=old effects
          WORD      wSpecial;
          BYTE      bGlobalVol;                             // global  volume [0;128]
          BYTE      bMasterVol;                             // master  volume [0;128]
          BYTE      bInitSpeed;                             // initial speed
          BYTE      bInitTempo;                             // initial tempo
          BYTE      bPanningSep;                            // panning separation
          BYTE      bZero;                                  // 00
          WORD      wMsgLen;                                // msg length
          DWORD     dwMsgOffset;                            // msg offset
          char      szUnused2[ 4];
          BYTE      bChanPan [64];                          // channel panning [0;64]
          BYTE      bChanVol [64];                          // channel volume  [0;64]
} IT_HEADER;                                                // 192 bytes

#pragma   pack()

/************************************************ CLASSES *************************************************************/

/**********************************************************************************************************************/
/*                                                IT                                                                  */
/**********************************************************************************************************************/

class     CIT       : public CSoundTrack
  {
// public methods

  public:

    // constructors & destructor

	  CIT                                 (void);
	 ~CIT                                 (void);

    void            vInit               (void);
    void            vClean              (void);

    // load & inits

    const char*     lpszGetExtension    (void);             // return file extension
    HRESULT         hrCheckSignature    (void);             // check the current file is an IT 
    HRESULT         hrInitVars          (void);             // init starting values + samples & patterns pointers

    // pattern data

    void            vGoToRow            (BYTE** lplpbPattern,const BYTE bRow);
    void            vUnpackRow          (BYTE*  lpbUnpacked,BYTE** lplpbNote);

    // get

    const char*     lpszGetTitle        (void) const;

// protected members

  protected:

    // file

    static const char
                    m_szExtension[_MAX_EXT];                // IT file extension
  };

/************************************************ INLINE **************************************************************/

/**********************************************************************************************************************/

#endif // !defined(AFX_IT_H__35F76920_BC62_11D3_9848_D0068C1C3643__INCLUDED_)
