/**********************************************************************************************************************/
/*                                                MOD CLASS                                                           */
/* 16/11/99, M : creation of the file                                                                                 */
/*               handles a MOD soundtrack                                                                             */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "MOD.h"
#include  "Global_err.h"
#include  "SoundTrack_err.h"

#include  "Debug.h"
#include  "Instrument.h"
#include  "Sample.inl"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/************************************************ STATIC MEMBERS ******************************************************/

const char CMOD::m_szExtension[_MAX_EXT] = ".MOD";          // MOD file extension

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

//*********************************************** constructors ********************************************************/
// 16/11/99, M
/**********************************************************************************************************************/

CMOD::CMOD()
  {
  _DTRCDEF_("CMOD::CMOD");
  //vInit();
  }

//*********************************************** destructor **********************************************************/
// 16/11/99, M
/**********************************************************************************************************************/

CMOD::~CMOD()
  {
  _DTRCDEF_("CMOD::~CMOD");
  //vClean();
  }

//*********************************************** vInit ***************************************************************/
// initialize members & resources
// 16/11/99, M
// in :
// out:
/**********************************************************************************************************************/

void CMOD::vInit()
  {
  _DTRCDEF_("CMOD::vInit");
  CSoundTrack::vInit();
  }

//*********************************************** vClean **************************************************************/
// release memory & resources
// 16/11/99, M
// in :
// out:
/**********************************************************************************************************************/

void CMOD::vClean()
  {
  _DTRCDEF_("CMOD::vClean");
  CSoundTrack::vClean();
  }

/**********************************************************************************************************************/
/*                                                LOAD & INITS                                                        */
/**********************************************************************************************************************/

//*********************************************** lpszGetExtension ****************************************************/
// return MOD extension
// 16/11/99, M
// in :
// out: MOD extension
/**********************************************************************************************************************/

const char* CMOD::lpszGetExtension() const
  {
  _DTRCDEF_("CMOD::lpszGetExtension");
  return(m_szExtension);
  }

//*********************************************** hrCheckSignature ****************************************************/
// check the current file is a MOD
// 16/11/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMOD::hrCheckSignature()
  {
  _DTRCDEF_("CMOD::hrCheckSignature");
  HRESULT hrError = CSoundTrack::hrCheckSignature();
  if(hrError) return(hrError);

  // compare MOD signature

  m_dwChannels = 0;
  MOD_HEADER* lpHeader = (MOD_HEADER*)m_lpbFile;

  if(strncmp( lpHeader->szMK,  "M.K.",4) == 0) m_dwChannels =  4;                                                       // M.K.
  if(strncmp( lpHeader->szMK,  "FLT" ,3) == 0) m_dwChannels =  *(lpHeader->szMK+3)-'0';                                 // FLTn
  if(strncmp( lpHeader->szMK+1,"CHN" ,3) == 0) m_dwChannels =  *(lpHeader->szMK  )-'0';                                 // nCHN
  if(strncmp( lpHeader->szMK+2,"CH"  ,2) == 0) m_dwChannels = (*(lpHeader->szMK  )-'0')*10 + *(lpHeader->szMK+1)-'0';   // nnCH
  
  if(!m_dwChannels)
    {
    _DWAR_("ERROR: not a MOD file");
    char buf[256];
    strcpy(buf,"unknown signature: ");
    strncat(buf,lpHeader->szMK,4);
//    AfxMessageBox(buf);
    return(m_hrError = _ERR_(_S3MERR_BADFORMAT_));
    }

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

//*********************************************** dwModC4Speed ********************************************************/
// remap MOD finetune on S3M C4Speed
// 16/11/99, M
// in : MOD finetune
// out: S3M C4Speed
/**********************************************************************************************************************/

DWORD CMOD::dwModC4Speed(const BYTE bFineTune)
  {
  _DTRCDEF_("CMOD::dwModC4Speed");

  switch(bFineTune & 0x0f)
    {
		case  0 : return 8363;
		case  1 : return 8413;
		case  2 : return 8463;
		case  3 : return 8529;
		case  4 : return 8581;
		case  5 : return 8651;
		case  6 : return 8723;
		case  7 : return 8757;
		case  8 : return 7895;
		case  9 : return 7941;
		case 10 : return 7985;
		case 11 : return 8046;
		case 12 : return 8107;
		case 13 : return 8169;
		case 14 : return 8232;
		case 15 : return 8280;
		default : return 8363;
    }
  }

//*********************************************** hrInitVars **********************************************************/
// init starting values + samples & patterns pointers
// 16/11/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMOD::hrInitVars()
  {
  _DTRCDEF_("CMOD::hrInitVars");
  HRESULT hrError = CSoundTrack::hrInitVars();
  if(hrError) return(hrError);

  // starting values

  MOD_HEADER* lpHeader = (MOD_HEADER*)m_lpbFile;

  m_bPatternSpeed0 = 6;
  m_bTempo0        = 125;
  m_bGlobalVolume0 = 64;
  m_bMasterVolume  = 48;
  m_boStereo0      = false;

  // balance

  if(!m_dwChannels) m_dwChannels  = 4;
  m_dwPattSize = m_dwChannels*64*4;                         // 64 rows, 4 bytes

  for(DWORD dwI = 0; dwI < m_dwChannels; dwI++)
    {
    m_boChanActive[dwI] = true;
    m_bPanPos0    [dwI] = 8;
    }

  // orders

  BYTE* lpbOrders = lpHeader->bOrderList;
  BYTE  bPattMax  = 0;
  DWORD dwPattMem;
  DWORD dwLastOrder;

  for(dwI = 0; dwI < _MAX_ORDERS_; dwI++) m_bOrderList[dwI] = 255;
  for(dwI = 0; dwI < 128;          dwI++) 
    {
    m_bOrderList[dwI] = *(lpbOrders+dwI);
    if(bPattMax < m_bOrderList[dwI]) bPattMax = m_bOrderList[dwI];

    if(m_bOrderList[dwI]) dwLastOrder = dwI;
    }
  for(dwI = dwLastOrder+1; dwI < 128; dwI++) m_bOrderList[dwI] = 255;

  m_bOrders = BYTE(dwLastOrder+1);
  bPattMax++;
  dwPattMem = bPattMax*m_dwPattSize;

  // samples

  DWORD dwSampStart = sizeof(MOD_HEADER) + dwPattMem;

  for(dwI = 0; dwI < _MAX_INSTRUMENTS_; dwI++) m_lpInstrument[dwI] = NULL;
  for(dwI = 0; dwI < 31;                dwI++)
    {
    // header

    MODSMP_HEADER* lpSMPHeader = &lpHeader->smpSample[dwI];

    CModSample   * lpSample = new CModSample;
    CInstrument  * lpInstr  = new CInstrument(lpSample);    // mono-sample instrument
    if(!lpSample || !lpInstr)
      {
      _DWAR_("ERROR: out of memory");
      m_boInitVars = false;
      return(m_hrError = _ERR_(_GLOBERR_OUTOFMEMORY_));
      }
    m_lpInstrument[dwI] = lpInstr;

    // data

    _SWAP_BYTES_(&lpSMPHeader->wLength);
    _SWAP_BYTES_(&lpSMPHeader->wLoopStart);
    _SWAP_BYTES_(&lpSMPHeader->wLoopLen);

    if(lpSMPHeader->wLength)
      {                                                     // there is a sample
      lpSample->vSetHeader (lpSMPHeader);                   // header   address
      lpSample->vSetDataRAM(m_lpbFile+dwSampStart);         // data RAM address
      lpSample->vSetC4Speed(dwModC4Speed(lpSMPHeader->bFineTune));
      }
    else
      {                                                     // no sample
      lpSample->vSetHeader (NULL);
      lpSample->vSetDataRAM(NULL);
      lpSample->vSetC4Speed(0);
      }

    // other variables

    if(lpSMPHeader->wLoopLen > 1)
      {                                                     // loop
      lpSample->vSetRepeat(lpSMPHeader->wLoopStart*2);
      lpSample->vSetLength(lpSMPHeader->wLength   *2);
      lpSample->vSetRepLen(lpSMPHeader->wLoopLen  *2);
      }
    else
      {
      lpSample->vSetLength(lpSMPHeader->wLength   *2);
      lpSample->vSetRepLen(0);
      }

    lpSample->vSet16Bits(false);
    lpSample->vSetVolume(lpSMPHeader->bVolume);

    // signed->unsigned conversion

    if(lpSample->dwGetLength())
      {
      BYTE* lpbData = lpSample->lpbGetDataRAM();
      DWORD dwLen   = lpSample->dwGetLength  ();

      for(DWORD dwSmp = 0; dwSmp < dwLen; dwSmp++)
        {
        *lpbData++ += 128;
        }
      }

    dwSampStart += lpSample->dwGetLength();
    }

  // patterns

  for(dwI = 0; dwI < _MAX_PATTERNS_; dwI++) m_lpbPattern[dwI] = NULL;
  for(dwI = 0; dwI < bPattMax;       dwI++) m_lpbPattern[dwI] = m_lpbFile + sizeof(MOD_HEADER) + m_dwPattSize*dwI;

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PATTERN DATA                                                        */
/**********************************************************************************************************************/

//*********************************************** vGoToRow ************************************************************/
// go to given row in packed pattern
// 16/11/99, M
// in : pattern start address,wanted row
// out: pattern row   address in lplpbPattern
/**********************************************************************************************************************/

void CMOD::vGoToRow(BYTE** lplpbPattern,const BYTE bRow)
  {
  _DTRCDEF_("CMOD::vGoToRow");
  if(!*lplpbPattern) return;

  *lplpbPattern += bRow*m_dwChannels*4;                     // 4 bytes by channel
  }

//*********************************************** vUnpackRow **********************************************************/
// unpack one row
// 16/11/99, M
// in : where to store unpacked row,packed data address (modified)
// out: 
/**********************************************************************************************************************/

void CMOD::vUnpackRow(BYTE*  lpbUnpacked,BYTE** lplpbNote)
  {
  _DTRCDEF_("CMOD::vUnpackRow");
  if(!*lplpbNote) return;
  BYTE* lpbNote = *lplpbNote;

  for(DWORD dwChan = 0; dwChan < m_dwChannels; dwChan++)
    {
    BYTE  bChannel = (BYTE)dwChan;
    BYTE* lpbChann = lpbUnpacked + bChannel*5;

    BYTE  bVal1    = *lpbNote++;
    BYTE  bVal2    = *lpbNote++;
    BYTE  bVal3    = *lpbNote++;
    BYTE  bVal4    = *lpbNote++;

    BYTE  bSmp     = BYTE ((bVal1 & 0xf0)       + (bVal3 >> 4));
    WORD  wPeriod  = WORD(((bVal1 & 0x0f) << 8) +  bVal2);
    BYTE  bCommand = BYTE(  bVal3 & 0x0f);
    BYTE  bInfo    = bVal4;
    BYTE  bVolume  = 0xff;

    // remap MOD periods on S3M notes

    BYTE  bNote    = 0xff;
    for(DWORD dwNote = 2*12; dwNote < 8*12; dwNote++)
      {
      if(wPeriod >= m_dwNoteFreq[dwNote])
        {
        bNote = BYTE(dwNote-2*12);                          // 2 octaves difference between MOD & S3M
        bNote = BYTE(((bNote/12) << 4) + (bNote % 12));     // octave+note
        break;
        }
      }

    // remap effects

    switch(bCommand)
      {
      case  0 : if(bInfo) bCommand =  10; break;            // arpeggio
      case  1 : bCommand =   6; break;                      // slide up
      case  2 : bCommand =   5; break;                      // slide down
      case  3 : bCommand =   7; break;                      // tone portamento
      case  4 : bCommand =   8; break;                      // vibrato
      case  5 : bCommand =  12; break;                      // tone portamento + vol slide
      case  6 : bCommand =  11; break;                      // vibrato         + vol slide
      case  7 : bCommand =  18; break;                      // tremolo
      case  8 : bCommand =   0; break;                      // unused
      case  9 : bCommand =  15; break;                      // sample offset
      case 10 : bCommand =   4; break;                      // vol slide
      case 11 : bCommand =   2; break;                      // jump to order
      case 12 : bCommand =   0; bVolume = bInfo; break;     // set volume
      case 13 : bCommand =   3; break;                      // break pattern
      case 15 : if(bInfo < 0x20) bCommand =  1;             // set speed
                else             bCommand = 20;             // set tempo
                break;
      case 14 : switch(bInfo >> 4)
                  {
                  case  0 : break;                                              // S0: unsupported
                  case  1 : bCommand =  6; bInfo |= 0xf0; break;                // fine slide up
                  case  2 : bCommand =  5; bInfo |= 0xf0; break;                // fine slide down
                  case  3 : break;                                              // S1: unsupported
                  case  4 : break;                                              // S3: unsupported
                  case  5 : break;                                              // S2: unsupported
                  case  6 : break;                                              // S4: unsupported
                  case  7 : break;                                              // set tremolo waveform
                  case  8 : bCommand = 19; break;                               // set panning
                  case  9 : bCommand = 17; bInfo &= 0x0f; break;                // retrig note
                  case 10 : bCommand =  4; bInfo  = BYTE((bInfo << 4)+0x0f);    // fine vol slide up
                            break;
                  case 11 : bCommand =  4; bInfo |= 0xf0; break;                // fine vol slide down
                  case 12 : bCommand = 19; break;                               // cut note
                  case 13 : break;                                              // SD: unsupported
                  case 14 : break;                                              // SE: unsupported
                  case 15 : break;                                              // SF: unsupported
                  }
                break;
      }

    //

    *(lpbChann+0)  = bNote;
    *(lpbChann+1)  = bSmp;
    *(lpbChann+2)  = bVolume;
    *(lpbChann+3)  = bCommand ? bCommand : BYTE(0xff);
    *(lpbChann+4)  = bCommand ? bInfo    : BYTE(0xff);
    }

  *lplpbNote = lpbNote;
  }

//*********************************************** lpszGetTitle ********************************************************/
// return song title
// 19/11/99, M
// in :
// out: title address
/**********************************************************************************************************************/

const char* CMOD::lpszGetTitle() const
  {
  ((MOD_HEADER*)m_lpbFile)->szMODName[20] = '\0';
  return((const char*)((MOD_HEADER*)m_lpbFile)->szMODName);
  }
