/**********************************************************************************************************************/
/*                                                MIXER CLASS                                                         */
/* 03/10/99, M : creation of the file                                                                                 */
/*               mixes 1 or more soundtracks + individual samples at different qualities                              */
/* 28/10/99, M : class reorg                                                                                          */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "Mixer.h"
#include  "Mixer_err.h"
#include  "Global_err.h"
#include  "SoundTrack_err.h"

#include  "Debug.h"
#include  "ParamsMixer.h"
#include  "SoundTrack.inl"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/************************************************ STATIC MEMBERS ******************************************************/

DWORD     CMixer::m_dwNbMixer      = 0;

DWORD*    CMixer::m_lpdwMixBuffer  = NULL;                  // mixing buffer
BYTE *    CMixer::m_lpbDMABuffer   = NULL;                  // final  buffer
BYTE *    CMixer::m_lpbPostTable   = NULL;                  // postprocessing  table
SBYTE*    CMixer::m_lpsbVolTable   = NULL;                  // volume  [0..64] table
SBYTE*    CMixer::m_lpsbPanTable   = NULL;                  // panning [0..15] table

DWORD     CMixer::m_dwNbParams     = 0;
DWORD     CMixer::m_dwMaxNbParams  = 0;

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// 03/01/99, M
/**********************************************************************************************************************/

CMixer::CMixer()
  {
  _DTRCDEF_("CMixer::CMixer");
  m_dwNbMixer++;
  vInit();
  }

/************************************************ destructor **********************************************************/
// 03/10/99, M
/**********************************************************************************************************************/

CMixer::~CMixer()
  {
  _DTRCDEF_("CMixer::~CMixer");
  m_dwNbMixer--;
  vClean();
  }

/************************************************ vInit ***************************************************************/
// initialize members & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CMixer::vInit()
  {
  _DTRCDEF_("CMixer::vInit");
  m_lptrkSong = NULL;
  m_hrError   = _ERR_(_OK_);

  // params mixer (1 by default)

  m_lpParamRoot = m_lpParamCur = m_lpParamLast = new CParamsMixer;
  m_dwNbParams  = 1;
  if(m_dwNbParams > m_dwMaxNbParams) m_dwMaxNbParams = m_dwNbParams;
  }

/************************************************ vClean **************************************************************/
// release memory & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CMixer::vClean()
  {
  _DTRCDEF_("CMixer::vClean");
  hrCleanBuffers();

  // params mixer linked list

  CParamsMixer* lpParams = m_lpParamRoot;
  CParamsMixer* lpNext;

  while(lpParams)
    {
    lpNext = lpParams->m_lpNext;
    lpParams->m_boPlay = false;                             // in case hrStopPlaying has not been called
    delete lpParams;
    lpParams = lpNext;
    }

  m_lpParamRoot = m_lpParamCur = m_lpParamLast = NULL;
  m_dwNbParams  = 0;
  }

/**********************************************************************************************************************/
/*                                                INITS                                                               */
/**********************************************************************************************************************/

/************************************************ hrInitBuffers *******************************************************/
// init static buffers & tables
// 10/10/99, M
// in : 16b?
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrInitBuffers(const bool bo16Bits)
  {
  _DTRCDEF_("CMixer::hrInitBuffers");
  if(!m_lptrkSong) return(m_hrError = _ERR_(_S3MERR_NOSONG_));

  if(!m_lpdwMixBuffer)
    {                                                       // else already done
    // allocate

    m_lpdwMixBuffer = new DWORD[_DMABUFSIZE_*2];
    m_lpbDMABuffer  = new BYTE [_DMABUFSIZE_*4];
    m_lpbPostTable  = new BYTE [_POSTBUFSIZE_];
    m_lpsbVolTable  = new SBYTE[_VOLBUFSIZE_];
    m_lpsbPanTable  = new SBYTE[_PANBUFSIZE_];

    if(!m_lpdwMixBuffer || !m_lpbDMABuffer || !m_lpbPostTable || !m_lpsbVolTable || !m_lpsbPanTable)
      {
      _DWAR_("ERROR: out of memory");
      hrCleanBuffers();
      return(m_hrError = _ERR_(_GLOBERR_OUTOFMEMORY_));
      }

    // build volume table

    for(DWORD dwVol = 0; dwVol <= 64; dwVol++) for(SDWORD sdwSmp = 0;sdwSmp <= 255; sdwSmp++)
      *(m_lpsbVolTable+sdwSmp+(dwVol << 8)) = BYTE(((sdwSmp-128)*dwVol) >> 6);

    // build postprocessing table

    BYTE bMaster = m_lptrkSong->bGetMasterVolume();
    if(bMaster < 0x10) bMaster = 0x10;

    DWORD dwC = 2048*16/bMaster;
    DWORD dwA = (2048-dwC) >> 1;
    DWORD dwB = dwA+dwC;
    DWORD dwX;

    for(dwX = 0;   dwX < dwA;  dwX++) *(m_lpbPostTable+dwX) = 0;
    for(dwX = dwA; dwX < dwB;  dwX++) *(m_lpbPostTable+dwX) = BYTE(((dwX-dwA) << 8) / (dwB-dwA));
    for(dwX = dwB; dwX < 2048; dwX++) *(m_lpbPostTable+dwX) = 255;

    // build panning table

    for(BYTE bPan = 0; bPan < 16; bPan++) for(SDWORD sdwSmp = 0; sdwSmp < 256; sdwSmp++)
      {
      SDWORD sdwVal = sdwSmp;
      if(sdwVal >= 128) sdwVal -= 256;
      *(m_lpsbPanTable+(bPan << 9)+(sdwSmp << 1)+0) = SBYTE(sdwVal*    bPan /15);         // left
      *(m_lpsbPanTable+(bPan << 9)+(sdwSmp << 1)+1) = SBYTE(sdwVal*(15-bPan)/15);         // right
      }
    }

  // init DMABuffer & MixBuffer to silence

  if(m_dwNbParams < 2)
    {
      if(!bo16Bits)
      {
      memset(m_lpbDMABuffer,0x80,_DMABUFSIZE_*4);

      BYTE* lpbMixBuffer = (BYTE*)m_lpdwMixBuffer;
      for(DWORD dwI = 0; dwI < _DMABUFSIZE_*2; dwI++)
        {
        *lpbMixBuffer++ = 0;
        *lpbMixBuffer++ = 4;                                // output = 1024
        *lpbMixBuffer++ = 0;
        *lpbMixBuffer++ = 0;
        }
      }

    else
      {
      memset(m_lpbDMABuffer, 0,_DMABUFSIZE_*4);
      memset(m_lpdwMixBuffer,0,_DMABUFSIZE_*4*2);
      }
    }

  // end

  return(_ERR_(_OK_));
  }

/************************************************ hrCleanBuffers ******************************************************/
// release static buffers & tables
// 10/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrCleanBuffers()
  {
  _DTRCDEF_("CMixer::hrCleanBuffers");
  if(m_dwNbMixer) return(_ERR_(_OK_));

  delete [] m_lpdwMixBuffer;
  delete [] m_lpbDMABuffer;
  delete [] m_lpbPostTable;
  delete [] m_lpsbVolTable;
  delete [] m_lpsbPanTable;

  m_lpdwMixBuffer = NULL;
  m_lpbDMABuffer  = NULL;
  m_lpbPostTable  = NULL;
  m_lpsbVolTable  = NULL;
  m_lpsbPanTable  = NULL;

  return(_ERR_(_OK_));
  }

/************************************************ hrSetBpm ************************************************************/
// set timing counters according to given tempo
// 10/10/99, M
// in : tempo
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrSetBPM(const BYTE bTempo)
  {
  _DTRCDEF_("CMixer::hrSetBpm");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  DWORD dwTempo    = (DWORD)bTempo*256;
  DWORD dwMixSpeed = m_lpParamCur->m_dwMixSpeed;

  dwMixSpeed *= 0x280;                                      // 5/2 *256
  dwMixSpeed /= dwTempo;

  m_lpParamCur->m_wTimerSpeed    = (WORD)dwMixSpeed;
  m_lpParamCur->m_wTimerSpeedRem = 0;

  return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PLAY                                                                */
/**********************************************************************************************************************/

/************************************************ hrStartPlaying ******************************************************/
// initialize & start music at given order
// 16/11/99, M
// in : 1st order to play,row in order,mix speed,stereo flag,16b flag,repeat song?,new params?,sound handle address
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrStartPlaying(const BYTE bOrder,const BYTE bRow,const DWORD dwMix,const bool boStereo,const bool bo16Bits,
                               const bool boRepeat,const bool boNew,DWORD* lpdwHandle)
  {
  _DTRCDEF_("CMixer::hrStartPlaying");
  if(!m_lptrkSong)  return(_ERR_(_S3MERR_NOSONG_));
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  CParamsSong* lpParamsSong;
  if( m_lptrkSong->hrStartPlaying(bOrder,bRow,boStereo,boRepeat,boNew,lpdwHandle,&lpParamsSong)) return(_ERR_(_S3MERR_MIXERSTART_));

  // multiple playing or sound ?

  if(boNew)
    {
    CParamsMixer* lpParamNew = new CParamsMixer;
    if(!lpParamNew) return(_ERR_(_GLOBERR_OUTOFMEMORY_));

    lpParamNew   ->m_lpPrev = m_lpParamLast;
    m_lpParamLast->m_lpNext = lpParamNew;
    m_lpParamLast           = lpParamNew;
    m_lpParamCur            = lpParamNew;

    m_dwNbParams++;
    if(m_dwNbParams > m_dwMaxNbParams) m_dwMaxNbParams = m_dwNbParams;
    }

  //

  m_lpParamCur->m_dwMixSpeed   = dwMix;
  m_lpParamCur->m_boMix16b     = bo16Bits;
  m_lpParamCur->m_lpParamsSong = lpParamsSong;

  HRESULT hrError;
  hrError = hrSetBPM(m_lptrkSong->bGetTempo());
  if(hrError) return(m_hrError = hrError);
  hrError = hrInitBuffers(bo16Bits);
  if(hrError) return(m_hrError = hrError);

  m_lpParamCur->m_boPlay = true;
  return(_ERR_(_OK_));
  }

/************************************************ hrStopPlaying *******************************************************/
// stop playing
// 16/11/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrStopPlaying()
  {
  _DTRCDEF_("CMixer::hrStopPlaying");
  if(!m_lptrkSong)  return(_ERR_(_S3MERR_NOSONG_));
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));
  if( m_lptrkSong->hrStopPlaying()) return(_ERR_(_S3MERR_MIXERSTOP_));
  m_lpParamCur->m_boPlay = false;

  // sound CParamsMixer ?

  if(m_lpParamCur != m_lpParamRoot)
    {
    if(m_lpParamCur->m_lpPrev) m_lpParamCur->m_lpPrev->m_lpNext = m_lpParamCur->m_lpNext;
    if(m_lpParamCur->m_lpNext) m_lpParamCur->m_lpNext->m_lpPrev = m_lpParamCur->m_lpPrev;
    if(m_lpParamCur == m_lpParamLast)             m_lpParamLast = m_lpParamCur->m_lpPrev;
    m_lpParamCur = NULL;
    m_dwNbParams--;
    }

  return(_ERR_(_OK_));
  }

/************************************************ vClearDMABuffer *****************************************************/
// clear DMA buffer before mixing - static
// 16/11/99, M
// in :
// out:
/**********************************************************************************************************************/

void CMixer::vClearDMABuffer()
  {
  _DTRCDEF_("CMixer::vClearDMABuffer");
  if(!m_lpbDMABuffer) return;
  memset(m_lpbDMABuffer,0,_DMABUFSIZE_*4);
  }

/************************************************ hrHandler ***********************************************************/
// handle an IRQ (fill buffer)
// 10/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrHandler()
  {
  _DTRCDEF_("CMixer::hrHandler");
  if(!m_lptrkSong) return(m_hrError = _ERR_(_S3MERR_NOSONG_));

  DWORD dwParams = 0;
  while(!hrSetCurParams(dwParams))
    {
    dwParams++;
    if(!m_lpParamCur->m_boPlay) continue;

    hrFillBuffer();

    // Copy

    DWORD* lpdwMixBuffer = m_lpdwMixBuffer;
    BYTE * lpbDMABuffer  = m_lpbDMABuffer;
    BYTE * lpbPostTable  = m_lpbPostTable;
    DWORD  dwDMABufSize  = _DMABUFSIZE_;
    WORD * lpwDMABuffer  = (WORD*)lpbDMABuffer;

    bool   boStereo      = m_lptrkSong->boGetStereo();

    // PostProc

    if(boStereo) dwDMABufSize += dwDMABufSize;

    if(!m_lpParamCur->m_boMix16b) while(dwDMABufSize--)
      {                                                       // 8b
      *lpbDMABuffer ++ += *(lpbPostTable + *lpdwMixBuffer);
      *lpdwMixBuffer++  = 1024;
      }

    else
      {
      if(!boStereo) while(dwDMABufSize)
        {                                                     // 16b mono
        SDWORD sdwMix     = *lpdwMixBuffer;
        sdwMix          >>= 3;
        *lpwDMABuffer ++ += (WORD)sdwMix;
        *lpdwMixBuffer++  = 0;

        sdwMix            = *lpdwMixBuffer;
        sdwMix          >>= 3;
        *lpwDMABuffer ++ += (WORD)sdwMix;
        *lpdwMixBuffer++  = 0;                                // skip next value
        dwDMABufSize     -= 2;
        }

      else while(dwDMABufSize)
        {                                                     // 16b stereo
        SDWORD sdwMix     = *lpdwMixBuffer;
        sdwMix          >>= 2;
        *lpwDMABuffer ++ += (WORD)sdwMix;
        *lpdwMixBuffer++  = 0;

        sdwMix            = *lpdwMixBuffer;
        sdwMix          >>= 2;
        *lpwDMABuffer ++ += (WORD)sdwMix;
        *lpdwMixBuffer++  = 0;                                // skip next value
        dwDMABufSize     -= 2;
        }
      }
    }

  return(_ERR_(_OK_));
  }

/************************************************ hrFillBuffer ********************************************************/
// fill a clean MixBuffer
// 10/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrFillBuffer()
  {
  _DTRCDEF_("CMixer::hrFillBuffer");
  if(!m_lptrkSong)  return(m_hrError = _ERR_(_S3MERR_NOSONG_));
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  m_lpParamCur->m_wBufLen = 0;                              // bytes chunk
  m_lpParamCur->m_wBufPos = 0;

  DWORD dwSpeed = m_lpParamCur->m_wTimerSpeedRem;
  bool  boNext  = true;
  if(dwSpeed) boNext = false;

  do
    {
    // NextChunk

    if(boNext)
      {
      dwSpeed = m_lpParamCur->m_wTimerSpeed;
      }
    boNext = true;

    // EndOfChunk

    m_lpParamCur->m_wTimerSpeedRem = 0;
    SDWORD sdwPos = m_lpParamCur->m_wBufPos;
    sdwPos += dwSpeed;
    sdwPos -= _DMABUFSIZE_;

    if(sdwPos >= 0)
      {
      dwSpeed -= sdwPos;
      m_lpParamCur->m_wTimerSpeedRem = (WORD)sdwPos;
      }

    // MixChunk

    m_lpParamCur->m_wBufLen = (WORD)dwSpeed;

    // MixLoop

    BYTE bChann = 4;

//    for(BYTE bChannel = 0+bChann-1; bChannel < _MAX_CHANNELS_*0+bChann; bChannel++)
    for(BYTE bChannel = 0; bChannel < _MAX_CHANNELS_; bChannel++)
      {
      if(!m_lptrkSong->boIsChanActive(bChannel)) continue;

      // MixActive

      hrMixTrack(bChannel,m_lpParamCur->m_wBufPos,m_lpParamCur->m_wBufLen);
      }

    // MixOk

    if(!m_lpParamCur->m_wTimerSpeedRem) 
      {
      BYTE bSaveTempo = m_lptrkSong->bGetTempo();

      /*if*/(m_lptrkSong->hrProcessEffects(m_lpParamCur->m_dwMixSpeed));// return(hrStopPlaying());

      BYTE bNewTempo  = m_lptrkSong->bGetTempo();
      if(bNewTempo != bSaveTempo) hrSetBPM(bNewTempo);
      }

    // FullBuffer

    m_lpParamCur->m_wBufPos += m_lpParamCur->m_wBufLen;
    }
  while(m_lpParamCur->m_wBufPos < _DMABUFSIZE_);

  //

  return(_ERR_(_OK_));
  }
 
/************************************************ hrMixTrack **********************************************************/
// mix one track into MixBuffer (output = 8/16bits)
// 17/10/99, M
// in : channel,buffer position,bytes to fill
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrMixTrack(const BYTE bChannel,const WORD wBufPos,const WORD wBufLen)
  {
  _DTRCDEF_("CMixer::hrMixTrack");
  if(!m_lptrkSong)  return(m_hrError = _ERR_(_S3MERR_NOSONG_));
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));
  S3M_CHANNEL* lpchanInfo = m_lptrkSong->lpGetChannel(bChannel);
  if(!lpchanInfo)  return(m_hrError = _ERR_(_S3MERR_BADCHANNEL_));

  DWORD  dwLen    = wBufLen;
  bool   boStereo = m_lptrkSong->boGetStereo();
  DWORD* lpdwMix  = m_lpdwMixBuffer + wBufPos*(1+boStereo);
  bool   bo16Bits = lpchanInfo->boSampl16b;
  bool   boPingP  = lpchanInfo->boPingPong;
//boPingP = false;
  bool   boLoopB  = lpchanInfo->boLoopBack;
  BYTE * lpbSampl = lpchanInfo->lpbDataRAM;
  if(!lpbSampl) return(_ERR_(_OK_));

  DWORD  dwVolume = lpchanInfo->bVol + lpchanInfo->bTremolo;
  DWORD  dwVolume8;
  if(dwVolume >= 128) dwVolume =  0;
  if(dwVolume >=  63) dwVolume = 63;
  dwVolume    *= m_lptrkSong->bGetGlobalVolume();
  dwVolume   >>= 6;                                         // clear some bits
  dwVolume8    = dwVolume << 8;                             // *256 for table access

  DWORD dwOffAdd  = lpchanInfo->dwOffsetAdd;
  DWORD dwError   = lpchanInfo->dwError;
  DWORD dwFreq    = lpchanInfo->dwGusFreq;
  DWORD dwFreqRem = lpchanInfo->dwGusRem;
  DWORD dwPanPos  = m_lptrkSong->bGetPanPos(bChannel);
  if(!m_lpParamCur->m_boMix16b) dwPanPos <<= 9;

  lpchanInfo->bParam2Set = 0;

  DWORD dwBufRep  = lpchanInfo->dwRepLen;
  DWORD dwRepeate = lpchanInfo->dwRepeate;
  DWORD dwCmpRepLen;
  if(bo16Bits)  dwBufRep   &= ~1;

  if(!dwBufRep) dwCmpRepLen = lpchanInfo->dwLen;
  else          dwCmpRepLen = dwRepeate + dwBufRep;

  DWORD dwSampVal,dwSampV2;
  DWORD dwFreq2 = (DWORD)-1;

  do
    {
    if(dwFreq2)
      {                                                     // dwOffAdd has changed (optim)
      // end or forward loop ?

      if(dwOffAdd >= dwCmpRepLen)
        {
        if(!dwBufRep) break;                                  // no loop

        if(!boPingP)  dwOffAdd -= dwBufRep;                   // forward   loop
        else          
          {                                                   // ping-pong loop
          boLoopB =  !boLoopB;
          DWORD dwOverFlow = dwOffAdd-dwCmpRepLen;
          dwOffAdd -= dwOverFlow;
          }
        }

      // backward loop

      if(boLoopB && (dwOffAdd <= dwRepeate))
        {
        boLoopB   = false;
        DWORD dwOverFlow = dwRepeate-dwOffAdd;
        dwOffAdd += dwOverFlow;
        }

      // interpolation part 1

      DWORD dwNext = dwOffAdd+1;
      if(bo16Bits)   dwNext++;
      if(dwNext   >= dwCmpRepLen)
        {
        if(dwBufRep) dwNext = dwRepeate;                      // loop
        else         dwNext = dwOffAdd;
        }

      // sample value

      if(!m_lpParamCur->m_boMix16b)
        {
        if(!bo16Bits) dwSampVal = *(lpbSampl+dwOffAdd);
        else          dwSampVal = *(lpbSampl+dwOffAdd+1);     // high weight
        if(!bo16Bits) dwSampV2  = *(lpbSampl+dwNext);
        else          dwSampV2  = *(lpbSampl+dwNext  +1);
        }
      else
        {
        if(!bo16Bits) dwSampVal = *(lpbSampl+dwOffAdd) << 8;
        else          dwSampVal = *(WORD*)(lpbSampl+dwOffAdd);
        if(!bo16Bits) dwSampV2  = *(lpbSampl+dwNext)   << 8;
        else          dwSampV2  = *(WORD*)(lpbSampl+dwNext);
        }
      }

    // interpolation

    DWORD dwSamp0 = dwSampVal;
    DWORD dwSamp1 = dwSampV2;
    BYTE  bError  = BYTE(dwError >> 24);

    dwSamp0  *= (256-bError);
    dwSamp1  *= bError;
    dwSamp0  += dwSamp1;
    dwSamp0 >>= 8;

    // volume

    SDWORD sdwSampVol;

    if(!m_lpParamCur->m_boMix16b)
      {
      sdwSampVol   = *(m_lpsbVolTable+dwVolume8+dwSamp0);
      }
    else
      {
      sdwSampVol   = dwSamp0 - (128 << 8);
      sdwSampVol  *= dwVolume;
      sdwSampVol >>= 6;
      }

    // stereo => panning

    if(boStereo)
      {
      SDWORD sdwPan;
      DWORD  dwSampVol;

      // panning table (8b)

      if(!m_lpParamCur->m_boMix16b)
        {
        dwSampVol   = (DWORD)sdwSampVol;
        dwSampVol  += dwSampVol;
        sdwPan      = *(m_lpsbPanTable+dwPanPos+dwSampVol+0); // left  panning
        *lpdwMix++ += sdwPan;

        sdwPan      = *(m_lpsbPanTable+dwPanPos+dwSampVol+1); // right panning
        *lpdwMix++ += sdwPan;
        }

      // 16b

      else
        {
        sdwPan      = sdwSampVol;
        sdwPan      *= dwPanPos;
        sdwPan     >>= 4;
        *lpdwMix++ += sdwPan;

        sdwPan      = sdwSampVol;
        sdwPan      *= 16-dwPanPos;
        sdwPan     >>= 4;
        *lpdwMix++ += sdwPan;
        }
      }

    // mono

    else
      {
      *lpdwMix++ += sdwSampVol;
      }

    // frequency

    DWORD dwOldErr = dwError;

    if(!boLoopB)
      {                                                     // forward loop
      dwError    += dwFreqRem;
      dwFreq2     = dwFreq;

      if(dwError < dwOldErr) dwFreq2++;
      dwOffAdd   += dwFreq2;
      if(bo16Bits)
        dwOffAdd += dwFreq2;
      }

    else
      {                                                     // backward loop
      dwError    -= dwFreqRem;
      dwFreq2     = dwFreq;

      if(dwError > dwOldErr) dwFreq2++;
      dwOffAdd   -= dwFreq2;
      if(bo16Bits)
        dwOffAdd -= dwFreq2;
      }

    dwLen--;
    }
  while(dwLen);

  // MixBye

  lpchanInfo->dwOffsetAdd = dwOffAdd;
  lpchanInfo->dwError     = dwError;
  lpchanInfo->boLoopBack  = boLoopB;

  return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PARAMS MIXER                                                        */
/**********************************************************************************************************************/

/************************************************ hrSetCurParams ******************************************************/
// set m_lpParamCur to nth params set in the list
// 03/12/99, M
// in : set number
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrSetCurParams(const DWORD dwNb)
  {
  _DTRCDEF_("CMixer::hrSetCurParams");
  if(!m_lptrkSong) return(m_hrError = _ERR_(_S3MERR_NOSONG_));

  // song

  HRESULT hrErr = m_lptrkSong->hrSetCurParams(dwNb);
  if(hrErr) return(hrErr);

  // mixer

  CParamsMixer* lpCur = m_lpParamRoot;
  DWORD         dwNum = dwNb;

  while(lpCur->m_lpNext && dwNum)
    {
    lpCur = lpCur->m_lpNext;
    dwNum--;
    }
  if(dwNum) return(_ERR_(_S3MERR_PARAMCUR_));               // not found, dwNb to big

  m_lpParamCur = lpCur;
  return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                SOUNDS                                                              */
/**********************************************************************************************************************/

/************************************************ hrSetPanning ********************************************************/
// change sound panning
// 03/12/99, M
// in : sound handle,new panning
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrSetPanning(const DWORD dwHandle,const BYTE bPan)
  {
  _DTRCDEF_("CMixer::hrSetPanning");
  if(!m_lptrkSong) return(m_hrError = _ERR_(_S3MERR_NOSONG_));
  return(m_lptrkSong->hrSetPanning(dwHandle,bPan));
  }

/************************************************ hrSetVolume *********************************************************/
// change sound volume
// 03/12/99, M
// in : sound handle,new volume
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CMixer::hrSetVolume(const DWORD dwHandle,const BYTE bVol)
  {
  _DTRCDEF_("CMixer::hrSetVolume");
  if(!m_lptrkSong) return(m_hrError = _ERR_(_S3MERR_NOSONG_));
  return(m_lptrkSong->hrSetVolume (dwHandle,bVol));
  }
