/**********************************************************************************************************************/
/*                                                MIXER CLASS                                                         */
/* 03/10/99, M : creation of the file                                                                                 */
/*               mixes 1 or more soundtracks + individual samples at different qualities                              */
/**********************************************************************************************************************/

#if !defined(AFX_MIXER_H__4C0D8F3C_79A8_11D3_9848_FD35403E9742__INCLUDED_)
#define AFX_MIXER_H__4C0D8F3C_79A8_11D3_9848_FD35403E9742__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif

/************************************************ INCLUDE *************************************************************/

#include  "GlobDef.h"

/************************************************ DEFINE **************************************************************/

// buffers'sizes

#define   _DMABUFSIZE_                  1024                // mono/stereo buffer size
#define   _MIXBUFSIZE_                  (_DMABUFSIZE_*4)    // same as _DMABUFSIZE_ but DWORDs
#define   _VOLBUFSIZE_                  (65*256)            // volume  [0..64] table size
#define   _POSTBUFSIZE_                 (2048*4)            // postprocessing  table size
#define   _PANBUFSIZE_                  (16*256*2)          // panning [0..15] table size

/************************************************ TYPEDEF & STRUCT ****************************************************/

/************************************************ CLASSES *************************************************************/

class     CParamsMixer;
class     CSoundTrack;

/**********************************************************************************************************************/
/*                                                Mixer                                                               */
/**********************************************************************************************************************/

class     CMixer
  {
// public methods

  public:

    // constructors & destructor

	  CMixer                              (void);
	 ~CMixer                              (void);

    void            vInit               (void);
    void            vClean              (void);

    // get

    static BYTE*    lpbGetDMABuffer     (void);

    // set

    void            vSetPlay            (const bool boPlay);
    void            vSetSong            (const CSoundTrack* lptrkSong);

    // buffers

    HRESULT         hrInitBuffers       (const bool bo16Bits);
    HRESULT         hrCleanBuffers      (void);

    // tempo

    HRESULT         hrSetBPM            (const BYTE bTempo);          // init Beats Per Minute (ticks speed) - effect Txx

    // play

    HRESULT         hrStartPlaying      (const BYTE bOrder,const BYTE bRow,const DWORD dwMix,const bool boStereo,const bool bo16bits,
                                         const bool boRepeat,const bool boNew=false,DWORD* lpdwHandle=NULL);
    HRESULT         hrStopPlaying       (void);                       // stop

    static void     vClearDMABuffer     (void);
    HRESULT         hrHandler           (void);                       // handle an IRQ (fill buffer)
    HRESULT         hrFillBuffer        (void);                       // fill a clean MixBuffer
    HRESULT         hrMixTrack          (const BYTE bChannel,const WORD wBufPos,const WORD wBufLen);

    // params mixer

    HRESULT         hrSetCurParams      (const DWORD dwNb);           // try to set params dwNb as the current ones (-1=last)

    // sounds

    HRESULT         hrSetPanning        (const DWORD dwHandle,const BYTE bPan);
    HRESULT         hrSetVolume         (const DWORD dwHandle,const BYTE bVol);

// protected members

  protected:

    // errors

    HRESULT         m_hrError;                              // error ID

    // song

    CSoundTrack*    m_lptrkSong;                            // soundtrack to play

    // params song

    CParamsMixer*   m_lpParamRoot;                          // linked list of playing params
    CParamsMixer*   m_lpParamCur;                           // current params set
    CParamsMixer*   m_lpParamLast;                          // last element in list

    static DWORD    m_dwNbParams;                           // nb elements in linked list
    static DWORD    m_dwMaxNbParams;                        // max nb elems

    // buffers

    static DWORD    m_dwNbMixer;

    static DWORD*   m_lpdwMixBuffer;                        // mixing buffer
    static BYTE *   m_lpbDMABuffer;                         // final  buffer
    static BYTE *   m_lpbPostTable;                         // postprocessing  table
    static SBYTE*   m_lpsbVolTable;                         // volume  [0..64] table
    static SBYTE*   m_lpsbPanTable;                         // panning [0..15] table
  };

/************************************************ INLINE **************************************************************/

//#include  "Mixer.inl"

/**********************************************************************************************************************/

#endif // !defined(AFX_MIXER_H__4C0D8F3C_79A8_11D3_9848_FD35403E9742__INCLUDED_)
