/**********************************************************************************************************************/
/*                                                SOUNDTRACK PLAYING PARAMETERS                                       */
/* 01/12/99, M : creation of the file                                                                                 */
/*               allows a soundtrack to be mixed more than once at the same time                                      */
/**********************************************************************************************************************/

#if !defined(AFX_PARAMSSONG_H__1BAB5500_A811_11D3_9848_CEC89456F541__INCLUDED_)
#define AFX_PARAMSSONG_H__1BAB5500_A811_11D3_9848_CEC89456F541__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif

/************************************************ INCLUDE *************************************************************/

/************************************************ DEFINE **************************************************************/

#define   _MAX_CHANNELS_                64                  // maximum number of channels

/************************************************ TYPEDEF & STRUCT ****************************************************/

// channel structure

#pragma   pack(1)

typedef   struct    {
          BYTE*     lpbDataRAM;                             // sample RAM address
          DWORD     dwOffsetAdd;                            // Oxx effect
          DWORD     dwError;                                // frequency error term
          DWORD     dwLen;                                  // sample length
          DWORD     dwRepeate;                              // sample loop start
          DWORD     dwRepLen;                               // sample loop length
          DWORD     dwFine;                                 // sample C4Speed
          DWORD     dwST3Freq;                              // note pitch
          DWORD     dwGusFreq;                              // real frequency
          DWORD     dwGusRem;                               // frequency remainder
          DWORD     dwPortTo;                               // Gxx effect
          BYTE      bSampl;                                 // sample number
          BYTE      bOldSampl;                              // previous sample
          BYTE      bVol;                                   // sample volume
          BYTE      bTremolo;                               // Rxy argument
          BYTE      bParam2Set;                             // 00001111 (sample-volume-frequency-note)
          BYTE      bLastNote;                              // last played note
          BYTE      bEffectComm;                            // command
          BYTE      bEffectInfo;                            // info
          BYTE      bOldEffect;                             // previous effect
          BYTE      bEffectTime;                            // "cut note" delay
          BYTE      bOldVolSlide;                           // last volume slide
          BYTE      bOldSlide;                              // last slide up/down
          BYTE      bOldPortTo;                             // last portamento to note
          BYTE      bOldVibrato;                            // last vibrato
          BYTE      bVibratoPos;                            // position in sinus table
          BYTE      bOldTremolo;                            // last tremolo
          BYTE      bTremoloPos;                            // position in sinus table
          BYTE      bOldOffset;                             // last offset(add)
          BYTE      bOldRetrig;                             // last retrig note
          BYTE      bOldArpeggio;                           // last arpeggio
          DWORD     dwArpVals [3];                          // half-tones for arpeggio
          BYTE      bArpCounter;                            // half-tone number
          bool      boSampl16b;                             // 16bits sample ?
          bool      boPingPong;                             // ping-pong looped sample ?
          bool      boLoopBack;                             // forward/backward loop direction
} S3M_CHANNEL;

#pragma   pack()

#define   _S3MCHANSIZE_                 sizeof(S3M_CHANNEL)

/************************************************ CLASSES *************************************************************/

/**********************************************************************************************************************/
/*                                                Params song                                                         */
/**********************************************************************************************************************/

class     CParamsSong
  {
// public methods

  public:

    // constructors & destructor

	  CParamsSong                         (void);
	 ~CParamsSong                         (void);       

    void            vInit               (void);
    void            vClean              (void);

// public members

  public:

    // flags

    bool            m_boInitParams;                         // vars initialized ? (with song initial values)

    // song

    bool            m_boStereo;                             // stereo OFF/ON
    BYTE            m_bGlobalVolume;                        // song volume multiplier (0-64)
    BYTE            m_bPatternSpeed;                        // number of ticks for each row
    BYTE            m_bTempo;                               // tick time (default=125 BPM)

    S3M_CHANNEL     m_chanInfo    [_MAX_CHANNELS_];         // channels infos
    BYTE            m_bPanPos     [_MAX_CHANNELS_];         // panning positions (0..15)

    // play

    BYTE            m_bPatternRow;                          // current row
    BYTE            m_bPlayingPattern;                      // current pattern number
    BYTE            m_bPatternCount;                        // pattern tick counter
    BYTE            m_bBreakData;                           // break (effect Cxx)
    BYTE            m_bSynchro;                             // demo synchro (Zxx)
    bool            m_boRepeat;                             // loop song ?
    bool            m_boPlaying;                            // processing ON

    WORD            m_wTimerCount;                          // IRQ   counter
    DWORD           m_dwRunningCount;                       // ticks counter
    float           m_fRunningTime;                         // time  counter (in seconds)
    BYTE*           m_lpbNote;                              // note packed data address

    BYTE            m_bUnpackedRow[_MAX_CHANNELS_][5];      // 1 row

    DWORD           m_dwHandle;                             // sound handle

    // linked list

    CParamsSong*    m_lpNext;
    CParamsSong*    m_lpPrev;
  };

/************************************************ INLINE **************************************************************/

//#include  "ParamsSong.inl"

/**********************************************************************************************************************/

#endif // !defined(AFX_PARAMSSONG_H__1BAB5500_A811_11D3_9848_CEC89456F541__INCLUDED_)
