/**********************************************************************************************************************/
/*                                                S3M CLASS                                                           */
/* 03/10/99, M : creation of the file                                                                                 */
/*               handles an S3M soundtrack                                                                            */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "S3M.h"
#include  "Global_err.h"
#include  "SoundTrack_err.h"

#include  "Debug.h"
#include  "Instrument.h"
#include  "Sample.inl"
#include  "SMP.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/************************************************ STATIC MEMBERS ******************************************************/

const char CS3M::m_szExtension[_MAX_EXT] = ".S3M";          // S3M file extension

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

//*********************************************** constructors ********************************************************/
// 03/01/99, M
/**********************************************************************************************************************/

CS3M::CS3M()
  {
  _DTRCDEF_("CS3M::CS3M");
  //vInit();
  }

//*********************************************** destructor **********************************************************/
// 03/10/99, M
/**********************************************************************************************************************/

CS3M::~CS3M()
  {
  _DTRCDEF_("CS3M::~CS3M");
  //vClean();
  }

//*********************************************** vInit ***************************************************************/
// initialize members & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CS3M::vInit()
  {
  _DTRCDEF_("CS3M::vInit");
  CSoundTrack::vInit();                       
  }

//*********************************************** vClean **************************************************************/
// release memory & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CS3M::vClean()
  {
  _DTRCDEF_("CS3M::vClean");
  CSoundTrack::vClean();
  }

/**********************************************************************************************************************/
/*                                                LOAD & INITS                                                        */
/**********************************************************************************************************************/

//*********************************************** lpszGetExtension ****************************************************/
// return S3M extension
// 04/10/99, M
// in :
// out: S3M extension
/**********************************************************************************************************************/

const char* CS3M::lpszGetExtension() const
  {
  _DTRCDEF_("CS3M::lpszGetExtension");
  return(m_szExtension);
  }

//*********************************************** hrCheckSignature ****************************************************/
// check the current file is an S3M
// 04/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CS3M::hrCheckSignature()
  {
  _DTRCDEF_("CS3M::hrCheckSignature");
  HRESULT hrError = CSoundTrack::hrCheckSignature();
  if(hrError) return(hrError);

  // compare S3M signature

  S3M_HEADER* lpHeader = (S3M_HEADER*)m_lpbFile;
  if(strncmp( lpHeader->szSCRM,"SCRM",4) != 0)
    {
    _DWAR_("ERROR: not an S3M file");
    return(m_hrError = _ERR_(_S3MERR_BADFORMAT_));
    }

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

//*********************************************** hrInitVars **********************************************************/
// init starting values + samples & patterns pointers
// 04/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CS3M::hrInitVars()
  {
  _DTRCDEF_("CS3M::hrInitVars");
  HRESULT hrError = CSoundTrack::hrInitVars();
  if(hrError) return(hrError);

  // starting values

  S3M_HEADER* lpHeader = (S3M_HEADER*)m_lpbFile;

  m_bPatternSpeed0 = lpHeader->bInitSpeed;
  m_bTempo0        = lpHeader->bInitTempo;
  m_bGlobalVolume0 = lpHeader->bGlobalVol;
  m_bMasterVolume  = BYTE(lpHeader->bMasterVol & 0x7F);
  m_boStereo0      =(lpHeader->bMasterVol & _S3MFLAG_STEREO_) != 0;

  // balance

  m_dwChannels     = 0;
  BYTE* lpbChanPan = m_lpbFile + 0x60 + lpHeader->wOrdNum + lpHeader->wInstNum*2 + lpHeader->wPattNum*2;

  for(DWORD dwI = 0; dwI < 32; dwI++)
    {                                                       // 32 channels
    BYTE bChanVal       = lpHeader->bChannels[dwI];
    m_boChanActive[dwI] = (bChanVal < _S3MFLAG_CHANDISABLED_);
    if(bChanVal != _S3MFLAG_CHANUNUSED_) 
      {
      m_dwChannels = dwI+1;

      if(m_boStereo0)
        {
        BYTE bPanPos = BYTE(3 + ((bChanVal & 8) != 0)*9);   // left=0x03 - right=0x0C

        if(lpHeader->bDefaultPan == _S3MFLAG_PANNING_)
          {
          BYTE bPos = *(lpbChanPan+dwI);
          if(bPos & 32) bPanPos = BYTE(bPos & 0xF);         // user defined balance
          }

        m_bPanPos0[dwI] = bPanPos;
        }

      else m_bPanPos0[dwI] = 8;
      }
    }

  // orders

  BYTE* lpbOrders = m_lpbFile + 0x60;

  for(dwI = 0; dwI < _MAX_ORDERS_;      dwI++) m_bOrderList[dwI] = 255;
  for(dwI = 0; dwI < lpHeader->wOrdNum; dwI++) m_bOrderList[dwI] = *(lpbOrders+dwI);

  m_bOrders = BYTE(lpHeader->wOrdNum);

  // samples

  WORD* lpwInst = (WORD*)(lpbOrders + lpHeader->wOrdNum);

  for(dwI = 0; dwI < _MAX_INSTRUMENTS_;  dwI++) m_lpInstrument[dwI] = NULL;
  for(dwI = 0; dwI < lpHeader->wInstNum; dwI++)
    {
    // header

    DWORD        dwParaPtr   = *(lpwInst+dwI) << 4;
    SMP_HEADER*  lpSMPHeader =  (SMP_HEADER*)(m_lpbFile + dwParaPtr);

    CSMP*        lpSample    = new CSMP;
    CInstrument* lpInstr     = new CInstrument(lpSample);   // mono-sample instrument
    if(!lpSample || !lpInstr)
      {
      _DWAR_("ERROR: out of memory");
      m_boInitVars = false;
      return(m_hrError = _ERR_(_GLOBERR_OUTOFMEMORY_));
      }
    m_lpInstrument[dwI] = lpInstr;

    // data

    if(lpSMPHeader->wMemSeg)
      {                                                     // there is a sample
      dwParaPtr = lpSMPHeader->wMemSeg << 4;
      lpSample->vSetHeader (lpSMPHeader);                   // header   address
      lpSample->vSetDataRAM(m_lpbFile+dwParaPtr);           // data RAM address
      lpSample->vSetC4Speed(lpSMPHeader->dwC4Speed);        // C4 speed
      }
    else
      {                                                     // no sample
      lpSample->vSetHeader (NULL);
      lpSample->vSetDataRAM(NULL);
      lpSample->vSetC4Speed(0);
      }

    // other variables

    if(lpSMPHeader->bFlags & _SMPFLAG_LOOP_)
      {                                                     // loop
      lpSample->vSetRepeat(lpSMPHeader->dwLoopStart);
      lpSample->vSetLength(lpSMPHeader->dwLength);
      lpSample->vSetRepLen(lpSMPHeader->dwLoopEnd - lpSMPHeader->dwLoopStart);
      }
    else
      {
      lpSample->vSetLength(lpSMPHeader->dwLength);
      lpSample->vSetRepLen(0);
      }

    lpSample->vSet16Bits((lpSMPHeader->bFlags & _SMPFLAG_16BITS_) != 0);
    lpSample->vSetVolume( lpSMPHeader->bVolume);
    }

  // patterns

  WORD* lpwPatt = lpwInst + lpHeader->wInstNum;

  for(dwI = 0; dwI < _MAX_PATTERNS_;     dwI++) m_lpbPattern[dwI] = NULL;
  for(dwI = 0; dwI < lpHeader->wPattNum; dwI++)
    {
    DWORD dwParaPtr = *(lpwPatt+dwI) << 4;

    if(dwParaPtr)
      {                                                     // not empty
      WORD wDataLength  = *(m_lpbFile+dwParaPtr);
      m_lpbPattern[dwI] = 2+m_lpbFile+dwParaPtr;            // skip pattern length
      }
    }

  // OK

  return(m_hrError = _ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PATTERN DATA                                                        */
/**********************************************************************************************************************/

//*********************************************** vGoToRow ************************************************************/
// go to given row in packed pattern
// 09/10/99, M
// in : pattern start address,wanted row
// out: pattern row   address in lplpbPattern
/**********************************************************************************************************************/

void CS3M::vGoToRow(BYTE** lplpbPattern,const BYTE bRow)
  {
  _DTRCDEF_("CS3M::vGoToRow");
  if(!*lplpbPattern) return;

  BYTE  bRowsToGo  = bRow;
  BYTE* lpbPattern = *lplpbPattern;

  while(bRowsToGo)
    {
    BYTE bVal = *lpbPattern++;

    if(bVal)
      {                                                     // 0 = empty row
      if(bVal &  32) lpbPattern += 2;                       // note & instr
      if(bVal &  64) lpbPattern ++;                         // volume
      if(bVal & 128) lpbPattern += 2;                       // command & info
      }

    else bRowsToGo--;                                       // next row
    }

  *lplpbPattern = lpbPattern;
  }

//*********************************************** vUnpackRow **********************************************************/
// unpack one row
// 09/10/99, M
// in : where to store unpacked row,packed data address (modified)
// out: 
/**********************************************************************************************************************/

void CS3M::vUnpackRow(BYTE*  lpbUnpacked,BYTE** lplpbNote)
  {
  _DTRCDEF_("CS3M::vUnpackRow");
  if(!*lplpbNote) return;
  BYTE* lpbNote = *lplpbNote;

  while(true)
    {
    BYTE bVal = *lpbNote++;

    if(bVal)
      {                                                     // 0 = empty row
      BYTE  bChannel = BYTE(bVal & 31);
      BYTE* lpbChann = lpbUnpacked + bChannel*5;

      if(bVal &  32) 
        {                                                   // note & instr
        *(lpbChann+0) = *lpbNote++;
        *(lpbChann+1) = *lpbNote++;
        }

      if(bVal & 64) *(lpbChann+2) = *lpbNote++;             // volume

      if(bVal & 128)
        {                                                   // command & info
        *(lpbChann+3) = *lpbNote++;
        *(lpbChann+4) = *lpbNote++;
        }
      }

    else
      {
      *lplpbNote = lpbNote;
      break;
      }
    }
  }

//*********************************************** lpszGetTitle ********************************************************/
// return song title
// 19/11/99, M
// in :
// out: title address
/**********************************************************************************************************************/

const char* CS3M::lpszGetTitle() const
  {
  ((S3M_HEADER*)m_lpbFile)->szS3MName[27] = '\0';
  return((const char*)((S3M_HEADER*)m_lpbFile)->szS3MName);
  }
