/**********************************************************************************************************************/
/*                                                S3M CLASS                                                           */
/* 03/10/99, M : creation of the file                                                                                 */
/*               handles an S3M soundtrack                                                                            */
/**********************************************************************************************************************/

#if !defined(AFX_S3M_H__4C0D8F3B_79A8_11D3_9848_FD35403E9742__INCLUDED_)
#define AFX_S3M_H__4C0D8F3B_79A8_11D3_9848_FD35403E9742__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif

/************************************************ INCLUDE *************************************************************/

#include  "SoundTrack.h"                                    // base class

/************************************************ DEFINE **************************************************************/

// header flags

#define   _S3MFLAG_VOLOPTIM_            8
#define   _S3MFLAG_AMIGALIMITS_         16
#define   _S3MFLAG_UNSIGNED_            2                   // unsigned sample
#define   _S3MFLAG_CHANDISABLED_        128                 // muted  channel
#define   _S3MFLAG_CHANUNUSED_          255                 // unused channel
#define   _S3MFLAG_STEREO_              128                 // stereo song
#define   _S3MFLAG_PANNING_             252                 // panning defined

/************************************************ TYPEDEF & STRUCT ****************************************************/

// S3M file header

#pragma   pack(1)

typedef   struct    {
          char      szS3MName[28];                          // song name (ASCIIZ)
          char      szID     [ 4];                          // 1A-16-xx-xx
          WORD      wOrdNum;                                // number of orders (even)
          WORD      wInstNum;                               // number of instruments
          WORD      wPattNum;                               // number of patterns
          WORD      wFlags;                                 // +8=0 vol optim; +16=amiga limits
          WORD      wVersion;                               // version of used tracker
          WORD      wSampSign;                              // 2=unsigned
          char      szSCRM   [ 4];                          // "SCRM" signature
          BYTE      bGlobalVol;                             // global  volume
          BYTE      bInitSpeed;                             // initial speed
          BYTE      bInitTempo;                             // initial tempo
          BYTE      bMasterVol;                             // master  volume
          BYTE      bUltraClick;                            // GUS click removal
          BYTE      bDefaultPan;                            // pan positions
          char      szUnused [ 8];
          WORD      wSpecial;
          BYTE      bChannels[32];                          // 0..15=pan, +128=disabled, 255=unused
} S3M_HEADER;                                               // 96 bytes

#pragma   pack()

/************************************************ CLASSES *************************************************************/

/**********************************************************************************************************************/
/*                                                S3M                                                                 */
/**********************************************************************************************************************/

class     CS3M      : public CSoundTrack
  {
// public methods

  public:

    // constructors & destructor

	  CS3M                                (void);
	 ~CS3M                                (void);

    void            vInit               (void);
    void            vClean              (void);

    // load & inits

    const char*     lpszGetExtension    (void) const;       // return file extension
    HRESULT         hrCheckSignature    (void);             // check the current file is an S3M
    HRESULT         hrInitVars          (void);             // init starting values + samples & patterns pointers

    // pattern data

    void            vGoToRow            (BYTE** lplpbPattern,const BYTE bRow);
    void            vUnpackRow          (BYTE*  lpbUnpacked,BYTE** lplpbNote);

    // get

    const char*     lpszGetTitle        (void) const;

// protected members

  protected:

    // file

    static const char
                    m_szExtension[_MAX_EXT];                // S3M file extension
  };

/************************************************ INLINE **************************************************************/

/**********************************************************************************************************************/

#endif // !defined(AFX_S3M_H__4C0D8F3B_79A8_11D3_9848_FD35403E9742__INCLUDED_)
