/**********************************************************************************************************************/
/*                                                SOUNDTRACK CLASS                                                    */
/* 03/10/99, M : creation of the file                                                                                 */
/*               handles a generic soundtrack (effects...)                                                            */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "SoundTrack.h"
#include  "SoundTrack_err.h"
#include  "Global_err.h"

#include  "Debug.h"
#include  "Instrument.h"
#include  "Sample.inl"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/************************************************ STATIC MEMBERS ******************************************************/

DWORD     CSoundTrack::m_dwNoteFreq[]   = {
            27392,25856,24384,23040,21696,20480,19328,18240,17216,16256,15360,14496,      // octave 0
            13696,12928,12192,11520,10848,10240, 9664, 9120, 8608, 8128, 7680, 7248,      // octave 1
             6848, 6464, 6096, 5760, 5424, 5120, 4832, 4560, 4304, 4064, 3840, 3624,      // octave 2
             3424, 3232, 3048, 2880, 2712, 2560, 2416, 2280, 2152, 2032, 1920, 1812,      // octave 3
             1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016,  960,  906,      // octave 4
              856,  808,  762,  720,  678,  640,  604,  570,  538,  508,  480,  453,      // octave 5
              428,  404,  381,  360,  339,  320,  302,  285,  269,  254,  240,  226,      // octave 6
              214,  202,  190,  180,  170,  160,  151,  143,  135,  127,  120,  113,      // octave 7
              107,  101,   95,   90,   85,   80,   75,   71,   67,   63,   60,   57,      // octave 8
               53,   50,   47,   45,   42,   40,   38,   36,   34,   32,   30,   28 };    // octave 9

BYTE      CSoundTrack::m_bSinusTable[]  = {
                0,   25,   50,   74,   98,  120,  142,  162,
              180,  197,  212,  225,  236,  244,  250,  254,
              255,  254,  250,  244,  236,  225,  212,  197,
              180,  162,  142,  120,   98,   74,   50,   25 };

DWORD     CSoundTrack::m_dwNbParams     = 0;
DWORD     CSoundTrack::m_dwMaxNbParams  = 0;
DWORD     CSoundTrack::m_dwHandleCount  = 1;

/**********************************************************************************************************************/
/*                                                CONSTRUCTORS / DESTRUCTOR                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// 03/01/99, M
/**********************************************************************************************************************/

CSoundTrack::CSoundTrack()
  {
  _DTRCDEF_("CSoundTrack::CSoundTrack");
  vInit();
  }

/************************************************ destructor **********************************************************/
// 03/10/99, M
/**********************************************************************************************************************/

CSoundTrack::~CSoundTrack()
  {
  _DTRCDEF_("CSoundTrack::~CSoundTrack");
  vClean();
  }

/************************************************ vInit ***************************************************************/
// initialize members & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CSoundTrack::vInit()
  {
  _DTRCDEF_("CSoundTrack::vInit");
  m_hrError    = _ERR_(_OK_);
  m_lpbFile    = NULL;
  m_dwFileSize = 0;

  m_boInitVars = false;

  // song defaults

  m_bPatternSpeed0 = 6;
  m_bTempo0        = 125;
  m_bGlobalVolume0 = 64;
  m_bMasterVolume  = 48;
  m_boStereo0      = true;
  m_dwChannels     = 0;

  // arrays

  for(DWORD dwI = 0; dwI < _MAX_ORDERS_; dwI++)
    m_bOrderList  [dwI] = 255;

  for(dwI = 0; dwI < _MAX_INSTRUMENTS_; dwI++)
    m_lpInstrument[dwI] = NULL;

  for(dwI = 0; dwI < _MAX_PATTERNS_; dwI++)
    m_lpbPattern  [dwI] = NULL;

  for(dwI = 0; dwI < _MAX_CHANNELS_; dwI++)
    {
    m_bPanPos0    [dwI] = 8;
    m_boChanActive[dwI] = false;
    }

  // params song (1 by default)

  m_lpParamRoot = m_lpParamCur = m_lpParamLast = new CParamsSong;
  m_dwNbParams  = 1;
  if(m_dwNbParams > m_dwMaxNbParams) m_dwMaxNbParams = m_dwNbParams;
  }

/************************************************ vClean **************************************************************/
// release memory & resources
// 03/10/99, M
// in :
// out:
/**********************************************************************************************************************/

void CSoundTrack::vClean()
  {
  _DTRCDEF_("CSoundTrack::vClean");

  // params song linked list

  CParamsSong* lpParams = m_lpParamRoot;
  CParamsSong* lpNext;

  while(lpParams)
    {
    lpNext = lpParams->m_lpNext;
    lpParams->m_boPlaying = false;                          // in case hrStopPlaying has not been called
    delete lpParams;
    lpParams = lpNext;
    }

  m_lpParamRoot = m_lpParamCur = m_lpParamLast = NULL;
  m_dwNbParams  = 0;

  //

  delete [] m_lpbFile;
  m_lpbFile = NULL;

  for(DWORD dwI = 0; dwI < _MAX_INSTRUMENTS_; dwI++)
    {
    delete m_lpInstrument[dwI];
    m_lpInstrument[dwI] = NULL;
    }
  }

/**********************************************************************************************************************/
/*                                                LOAD & INITS                                                        */
/**********************************************************************************************************************/

/************************************************ hrUnload ************************************************************/
// free file memory
// 08/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrUnload()
  {
  _DTRCDEF_("CSoundTrack::hrUnload");
  vClean();
  vInit ();
  return(_ERR_(_OK_));
  }

/************************************************ hrLoad **************************************************************/
// load file from disk
// 04/10/99, M
// in : file name
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrLoad(const char* lpszName)
  {
  _DTRCDEF_("CSoundTrack::hrLoad");
  hrUnload();

  // replace or add extension

  char  szDrive[_MAX_DRIVE];
  char  szDir  [_MAX_DIR];
  char  szName [_MAX_FNAME];
  char  szExt  [_MAX_EXT];
  char  szFile [_MAX_PATH];
  const char*  lpszExt;

  lpszExt = lpszGetExtension();
  _splitpath(lpszName,szDrive,szDir,szName,szExt);
  _makepath (  szFile,szDrive,szDir,szName,lpszExt ? lpszExt : szExt);

  // open file

  FILE* lpfileF = fopen(szFile,"rb");                       // read, binary file
  if(!lpfileF)
    {
    _DWAR_("ERROR opening file " << szFile);
    return(m_hrError = _ERR_(_GLOBERR_FOPEN_));
    }

  // get file size

  if(fseek(lpfileF,0L,SEEK_END))                            // go to end of file
    {
    _DWAR_("ERROR calling fseek");
    fclose(lpfileF);
    return(m_hrError = _ERR_(_GLOBERR_FSEEK_));
    }

  DWORD dwPos = ftell(lpfileF);                             // read position
  fseek(lpfileF,0,SEEK_SET);                                // go back to start of file

  if(dwPos == -1L)
    {
    _DWAR_("ERROR calling ftell");
    fclose(lpfileF);
    return(m_hrError = _ERR_(_GLOBERR_FTELL_));
    }

  // get some memory

  BYTE* lpbFile = new BYTE[dwPos];
  if(!lpbFile)
    {
    _DWAR_("ERROR: out of memory");
    fclose(lpfileF);
    return(m_hrError = _ERR_(_GLOBERR_OUTOFMEMORY_));
    }

  // read file

  DWORD dwRead = fread(lpbFile,1,dwPos,lpfileF);
  if(dwRead < dwPos)
    {
    _DWAR_("ERROR calling fread");
    fclose(lpfileF);
    delete [] lpbFile;
    return(m_hrError = _ERR_(_GLOBERR_FREAD_));
    }

  // OK

  m_lpbFile    = lpbFile;
  m_dwFileSize = dwPos;

  fclose(lpfileF);
  return(m_hrError = _ERR_(_OK_));
  }

/************************************************ lpszGetExtension ****************************************************/
// return file extension
// 04/10/99, M
// in :
// out: NULL (no default extension => don't modify)
/**********************************************************************************************************************/

const char* CSoundTrack::lpszGetExtension() const
  {
  _DTRCDEF_("CSoundTrack::lpszGetExtension");
  return(NULL);
  }

/************************************************ hrCheckSignature ****************************************************/
// check the current file type
// 04/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrCheckSignature()
  {
  _DTRCDEF_("CSoundTrack::hrCheckSignature");
  if(!m_lpbFile)
    {
    _DWAR_("ERROR: no file loaded");
    return(m_hrError = _ERR_(_S3MERR_NOFILELOADED_));
    }

  return(m_hrError = _ERR_(_OK_));
  }

/************************************************ hrInitVars **********************************************************/
// init starting values + samples & patterns pointers
// 04/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrInitVars()
  {
  _DTRCDEF_("CSoundTrack::hrInitVars");
  if(!m_lpbFile)
    {
    _DWAR_("ERROR: no file loaded");
    return(m_hrError = _ERR_(_S3MERR_NOFILELOADED_));
    }

  if(!m_lpParamRoot)
    {
    _DWAR_("ERROR: no default CParamsSong");
    return(m_hrError = _ERR_(_S3MERR_NOPARAMSONG_));
    }

  if(m_boInitVars)
    {
    _DWAR_("ERROR: inits already done");
    return(m_hrError = _ERR_(_S3MERR_INITDONE_));
    }

  m_boInitVars = true;
  return(m_hrError = _ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PLAY                                                                */
/**********************************************************************************************************************/

/************************************************ hrStartPlaying ******************************************************/
// initialize & start music at given order
// 08/10/99, M - modified 02/12/99 for sounds
// in : 1st order to play,row in order,stereo flag,repeat song?,new CParamsSong?,sound handle address,SongParams address
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrStartPlaying(const BYTE bOrder,const BYTE bRow,const bool boStereo,const bool boRepeat,const bool boNew,
                                    DWORD* lpdwHandle,CParamsSong** lplpParams)
  {
  _DTRCDEF_("CSoundTrack::hrStartPlaying");
  if(!m_boInitVars)
    {
    _DWAR_("ERROR: no file ready to be played");
    return(m_hrError = _ERR_(_S3MERR_NOINITDONE_));
    }
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  // multiple playing or sound ?

  if(boNew)
    {
    CParamsSong* lpParamNew = new CParamsSong;
    if(!lpParamNew) return(_ERR_(_GLOBERR_OUTOFMEMORY_));

    lpParamNew   ->m_lpPrev = m_lpParamLast;
    m_lpParamLast->m_lpNext = lpParamNew;
    m_lpParamLast           = lpParamNew;
    m_lpParamCur            = lpParamNew;

    m_dwNbParams++;
    if(m_dwNbParams > m_dwMaxNbParams) m_dwMaxNbParams = m_dwNbParams;
    }

  //

  m_lpParamCur->m_boStereo        = boStereo;               // or m_boStereo0
  m_lpParamCur->m_bGlobalVolume   = m_bGlobalVolume0;
  m_lpParamCur->m_bPatternSpeed   = m_bPatternSpeed0;
  m_lpParamCur->m_bTempo          = m_bTempo0;

  m_lpParamCur->m_bPatternRow     = 64;
  m_lpParamCur->m_bPlayingPattern = bOrder;
  m_lpParamCur->m_bPatternCount   = 1;
  m_lpParamCur->m_bBreakData      = bRow;
  m_lpParamCur->m_bSynchro        = 0;
  m_lpParamCur->m_boRepeat        = boRepeat;
  m_lpParamCur->m_dwRunningCount  = 0;
  m_lpParamCur->m_fRunningTime    = 0.f;

  memcpy(m_lpParamCur->m_bPanPos,m_bPanPos0,sizeof(m_bPanPos0));
  memset(m_lpParamCur->m_chanInfo,0,sizeof(m_lpParamCur->m_chanInfo));

  m_lpParamCur->m_boInitParams = true;
  m_lpParamCur->m_boPlaying    = true;

  //

  if(!m_lpParamCur->m_dwHandle)
    m_lpParamCur->m_dwHandle = m_dwHandleCount++;
  if(lplpParams) *lplpParams = m_lpParamCur;                // return CParamsSong
  if(lpdwHandle) *lpdwHandle = m_lpParamCur->m_dwHandle;    // return sound handle

  return(_ERR_(_OK_));
  }

/************************************************ hrStopPlaying *******************************************************/
// stop playing
// 08/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrStopPlaying()
  {
  _DTRCDEF_("CSoundTrack::hrStopPlaying");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));
  m_lpParamCur->m_boPlaying = false;

  // sound CParamsSong ?

  if(m_lpParamCur != m_lpParamRoot)
    {
    if(m_lpParamCur->m_lpPrev) m_lpParamCur->m_lpPrev->m_lpNext = m_lpParamCur->m_lpNext;
    if(m_lpParamCur->m_lpNext) m_lpParamCur->m_lpNext->m_lpPrev = m_lpParamCur->m_lpPrev;
    if(m_lpParamCur == m_lpParamLast)             m_lpParamLast = m_lpParamCur->m_lpPrev;
    m_lpParamCur = NULL;
    m_dwNbParams--;
    }

  return(_ERR_(_OK_));
  }

/************************************************ hrProcessEffects ****************************************************/
// patterns & effects management
// 09/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrProcessEffects(const DWORD dwMix)
  {
  _DTRCDEF_("CSoundTrack::hrProcessEffects");
  if(!m_lpParamCur)              return(_ERR_(_S3MERR_PARAMCUR_));
  if(!m_lpParamCur->m_boPlaying) return(_ERR_(_OK_));
  
  // ticks

  m_lpParamCur->m_wTimerCount = 0;                                    // next time : send notes
  m_lpParamCur->m_dwRunningCount++;                                   // music counter (synchro)
  m_lpParamCur->m_fRunningTime += 2.5f/m_lpParamCur->m_bTempo;        // time  counter (seconds)
  m_lpParamCur->m_bPatternCount--;                                    // one more tick

  if(!m_lpParamCur->m_bPatternCount)
    {                                                                 // row finished
    // DoPattern

    m_lpParamCur->m_bPatternCount = m_lpParamCur->m_bPatternSpeed;
                                                                      // reset row duration
    if(m_lpParamCur->m_bPatternRow >= 64)
      {                                                               // end of current pattern
      // NextOrder

      BYTE  bPattern = m_bOrderList[m_lpParamCur->m_bPlayingPattern];
      while(bPattern >= 254)
        {
        if(bPattern == 255)
          {                                                           // 255 = end
          if(!m_lpParamCur->m_bPlayingPattern || !m_lpParamCur->m_boRepeat) 
            {
            hrStopPlaying();
            return(_ERR_(_S3MERR_FINISHED_));
            }
          m_lpParamCur->m_bPlayingPattern = 0;                        // restart
          }

        else
          {                                                           // 254 = skip
          m_lpParamCur->m_bPlayingPattern++;
          }

        bPattern = m_bOrderList[m_lpParamCur->m_bPlayingPattern];
        }

      // NextOrderOk

      BYTE* lpbPattern = m_lpbPattern[bPattern];
      m_lpParamCur->m_bPatternRow = m_lpParamCur->m_bBreakData;
      m_lpParamCur->m_bBreakData  = 0;                                // row where to begin

      if(m_lpParamCur->m_bPatternRow && lpbPattern) vGoToRow(&lpbPattern,m_lpParamCur->m_bPatternRow);

      // SaveNote

      m_lpParamCur->m_lpbNote = lpbPattern;                           // save row address
      m_lpParamCur->m_bPlayingPattern++;
      }
    
    // NoNextPattern

    m_lpParamCur->m_bPatternRow++;
    memset(m_lpParamCur->m_bUnpackedRow,0xff,sizeof(m_lpParamCur->m_bUnpackedRow));
    if(m_lpParamCur->m_lpbNote) vUnpackRow((BYTE*)&(m_lpParamCur->m_bUnpackedRow),&(m_lpParamCur->m_lpbNote));
    
    // PattLoop

    for(BYTE bChannel = 0; bChannel < _MAX_CHANNELS_; bChannel++)
      {
      if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));             // because of hrStopPlaying

      if(m_boChanActive[bChannel])
        {
        // SoundChannel

        BYTE bOctNote = m_lpParamCur->m_bUnpackedRow[bChannel][0];    // octave-note
        BYTE bOct;
        BYTE bNote;

        if(bOctNote == 254)
          {                                                           // ^^
          m_lpParamCur->m_chanInfo[bChannel].dwGusFreq  = 0;          // no sound
          m_lpParamCur->m_chanInfo[bChannel].dwGusRem   = 0;          // no sound
          m_lpParamCur->m_chanInfo[bChannel].bParam2Set = 2;          // new frequency
          }
        else
          {
          // NotCutOff

          if(bOctNote != 255)                                         // 0xff = empty note
            {
            bNote  = BYTE(bOctNote & 0x0f);
            bOct   = BYTE(bOctNote >> 4);
            bNote += BYTE(bOct*12);                                   // note number (0..95)
            m_lpParamCur->m_chanInfo[bChannel].bLastNote  = bNote;    // save for arpeggio
            DWORD dwPeriod = m_dwNoteFreq[bNote];
            m_lpParamCur->m_chanInfo[bChannel].dwPortTo   = dwPeriod;
            m_lpParamCur->m_chanInfo[bChannel].bParam2Set = 2;        // new freq

            BYTE bComm = m_lpParamCur->m_bUnpackedRow[bChannel][3];   // command
            if((bComm != 7) && (bComm != 12))                         // Gxx (PortToTone) or Lxy (Vibrato+Vol slide)
              {
              m_lpParamCur->m_chanInfo[bChannel].bParam2Set  = 15;    // all is new
              m_lpParamCur->m_chanInfo[bChannel].bVibratoPos = 0;     // reset vibrato & tremolo
              m_lpParamCur->m_chanInfo[bChannel].bTremoloPos = 0;
              m_lpParamCur->m_chanInfo[bChannel].bTremolo    = 0;
              m_lpParamCur->m_chanInfo[bChannel].dwOffsetAdd = 0;
              m_lpParamCur->m_chanInfo[bChannel].dwError     = 0;
              m_lpParamCur->m_chanInfo[bChannel].boLoopBack  = false; // loop initial direction = forward
              }
            }
          }

        // NoNewNote

        CSample* lpSample = NULL;
        BYTE     bSample  = m_lpParamCur->m_bUnpackedRow[bChannel][1];// sample

        if((bSample != 0) && (bSample != 255))                        // 0x00 or 0xff : no sample
          {
          m_lpParamCur->m_chanInfo[bChannel].bSampl = bSample;
          if(m_lpInstrument[bSample-1])
            lpSample = m_lpInstrument[bSample-1]->lpSelectSample(bNote);

          if(lpSample)
            {
            m_lpParamCur->m_chanInfo[bChannel].bVol = lpSample->bGetVolume();
            m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;       // new volume
            }
          }

        // NoNewSample

        if(m_lpParamCur->m_chanInfo[bChannel].bParam2Set & 1)
          {                                                           // new note
          bSample = BYTE(m_lpParamCur->m_chanInfo[bChannel].bSampl-1);// 0..98
          if((bSample != 255) && m_lpInstrument[bSample])
            {
            lpSample = m_lpInstrument[bSample]->lpSelectSample(bNote);
            if(lpSample)
              {
              m_lpParamCur->m_chanInfo[bChannel].lpbDataRAM = lpSample->lpbGetDataRAM();
              m_lpParamCur->m_chanInfo[bChannel].dwFine     = lpSample->dwGetC4Speed ();
              m_lpParamCur->m_chanInfo[bChannel].dwLen      = lpSample->dwGetLength  () -1;
              m_lpParamCur->m_chanInfo[bChannel].dwRepeate  = lpSample->dwGetRepeat  ();
              m_lpParamCur->m_chanInfo[bChannel].dwRepLen   = lpSample->dwGetRepLen  ();
              m_lpParamCur->m_chanInfo[bChannel].boSampl16b = lpSample->boGet16Bits  ();
              m_lpParamCur->m_chanInfo[bChannel].boPingPong = lpSample->boGetPingPong();
              m_lpParamCur->m_chanInfo[bChannel].boLoopBack = false;  // loop initial direction = forward
              }
            }
          }

        // TstNewFreq

        if(m_lpParamCur->m_chanInfo[bChannel].bParam2Set & 2)
          {                                                           // new freq
          DWORD dwRate = m_lpParamCur->m_chanInfo[bChannel].dwPortTo;
          DWORD dwFine = m_lpParamCur->m_chanInfo[bChannel].dwFine;

          if(dwFine)
            {
            dwRate *= _MIDC_RATE_;
            dwRate /= dwFine;
            }
          else dwRate = 0;                                            // pb... sample not loaded ?

          m_lpParamCur->m_chanInfo[bChannel].dwPortTo = dwRate;

          if(m_lpParamCur->m_chanInfo[bChannel].bParam2Set & 1)
            {                                                         // new note
            m_lpParamCur->m_chanInfo [bChannel].dwST3Freq = dwRate;
            hrCalcuFreq(bChannel,dwRate,dwMix);
            }
          }

        // LoadVolume

        BYTE bVolume = m_lpParamCur->m_bUnpackedRow[bChannel][2];     // volume

        if(bVolume != 255)                                            // 0xFF = no volume
          {
          m_lpParamCur->m_chanInfo[bChannel].bVol        = bVolume;
          m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;
          }

        // LoadEffect

        BYTE bOldCommand = m_lpParamCur->m_chanInfo[bChannel].bEffectComm;
        BYTE bOldInfo    = m_lpParamCur->m_chanInfo[bChannel].bEffectInfo;
        m_lpParamCur->m_chanInfo[bChannel].bOldEffect = bOldCommand;

        BYTE bCommand = m_lpParamCur->m_bUnpackedRow[bChannel][3];    // command
        BYTE bInfo    = m_lpParamCur->m_bUnpackedRow[bChannel][4];    // info

        if((bOldCommand == 10) || (bCommand == 10))
          {                                                           // old or new = arpeggio ?
          DWORD dwArpVal = m_lpParamCur->m_chanInfo[bChannel].dwArpVals[0];
          m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwArpVal;    // continue old arpeggio

          DWORD dwFine = m_lpParamCur->m_chanInfo[bChannel].dwFine;

          if(dwFine)
            {
            dwArpVal *= _MIDC_RATE_;
            dwArpVal /= dwFine;
            }
          else dwArpVal = 0;

          hrCalcuFreq(bChannel,dwArpVal,dwMix);
          m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;         // new freq
          }

        // OldEffVib

        if((bOldCommand == 8) || (bOldCommand == 21) || (bOldCommand == 11))
          {                                                           // old = vibrato or fine vibrato or vib+vol slide
          if((bCommand != 8) && (bCommand != 21) && (bCommand != 11))
            {                                                         // new = no vibrato
            if(!(m_lpParamCur->m_chanInfo[bChannel].bParam2Set & 2))
              {                                                       // freq not already set
              DWORD dwFreq = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
              hrCalcuFreq(bChannel,dwFreq,dwMix);                     // let's finish the vibrato
              m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;     // old note freq
              }
            }
          }

        // Jump2Effect

        if(bCommand > 26)                                             // > 'Z' : unknown effect
          m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;
        else
          {
          m_lpParamCur->m_chanInfo[bChannel].bEffectComm = bCommand;
          m_lpParamCur->m_chanInfo[bChannel].bEffectInfo = bInfo;

          // PattJumps

          hrPattEffects(bChannel,bCommand,bInfo,dwMix);
          }
        }

      // NoPattEffect
      }
    }

  else
    {
    // DoEffects

    for(BYTE bChannel = 0; bChannel < _MAX_CHANNELS_; bChannel++)
      {
      if(m_boChanActive[bChannel])
        {
        // DoChannel

        m_lpParamCur->m_chanInfo[bChannel].bEffectTime--;             // one more tick

        m_lpParamCur->m_chanInfo[bChannel].bArpCounter++;             // rotate ArpCounter : 0,1,2
        if(m_lpParamCur->m_chanInfo[bChannel].bArpCounter >= 3) 
          m_lpParamCur->m_chanInfo[bChannel].bArpCounter = 0;

        BYTE bCommand = m_lpParamCur->m_chanInfo[bChannel].bEffectComm;
        BYTE bInfo    = m_lpParamCur->m_chanInfo[bChannel].bEffectInfo;

        // EffectJumps

        hrTickEffects(bChannel,bCommand,bInfo,dwMix);

        // NoEffect
        }
      }
    }

  return(_ERR_(_OK_));
  }

/************************************************ hrPattEffects *******************************************************/
// process pattern effects
// 13/10/99, M
// in : channel,effect,info
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrPattEffects(const BYTE bChannel,const BYTE bCommand,const BYTE bInfo,const DWORD dwMix)
  {
  _DTRCDEF_("CSoundTrack::hrPattEffects");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  switch(bCommand)
    {
    // Axx : set speed
    case  1 : if(bInfo) m_lpParamCur->m_bPatternSpeed = m_lpParamCur->m_bPatternCount = bInfo;
              break;

    // Bxx : jump to order
    case  2 : if(!m_lpParamCur->m_boRepeat && (bInfo <= m_lpParamCur->m_bPlayingPattern)) return(hrStopPlaying());
              m_lpParamCur->m_bPatternRow     = 64;                   // current pattern ends              
              m_lpParamCur->m_bPlayingPattern = bInfo;                // new order
              break;

    // Cxx : break pattern
    case  3 : if(bInfo <= 0x63)
                {
                BYTE bLow  = BYTE(bInfo & 0xf);
                if(bLow > 9) break;                                   // not a digit
                BYTE bHigh = BYTE(bInfo >> 4);

                m_lpParamCur->m_bBreakData  = BYTE(bHigh*10+bLow);    // 1st row to play
                m_lpParamCur->m_bPatternRow = 64;                     // current pattern ends
                }
              break;

    // DFy & DxF : fine vol slide
    case  4 : {
                BYTE bValue = bInfo;                                  // D00 => use old value
                if( !bValue)  bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide;
                m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide = bValue;

                BYTE bLow  = BYTE(bValue & 0xf);
                BYTE bHigh = BYTE(bValue >> 4);

                if(bLow == 0xf)
                  {                                                   // fine vol up
                  if(bHigh == 0) break;                               // D0F: do nothing
                  bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bLow += bHigh;
                  if(bLow >= 0x40) bLow = 0x3f;                       // max=63
                  }
                else if(bHigh == 0xf)
                  {                                                   // fine vol down
                  if(bLow == 0) break;                                // DF0: do nothing
                  bHigh  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bHigh -= bLow;
                  if(bHigh >= 0x80) bHigh = 0;                        // min=0
                  bLow = bHigh;
                  }
                else break;

                // FineVUok
                m_lpParamCur->m_chanInfo[bChannel].bVol        = bLow;
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;   // new volume
                m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;   // clear effect
              }
              break;

    // EFx & EEx : fine slide down
    case  5 : {
                BYTE bValue = bInfo;
                if( !bValue)  bValue = m_lpParamCur->m_chanInfo[bChannel].bOldSlide;
                m_lpParamCur->m_chanInfo[bChannel].bOldSlide = bValue;
                if(bValue < 0xe0) break;                              // not a fine slide

                BYTE bLow  = BYTE(bValue & 0xf);
                BYTE bHigh = BYTE(bValue >> 4);
                if(bHigh  != 0xe) bLow <<= 2;                         // fine = 4*extra fine

                DWORD dwFreq = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                dwFreq += bLow;
                if(dwFreq > m_dwNoteFreq[0])
                   dwFreq = m_dwNoteFreq[0];                          // cannot go below C0 note
                m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwFreq;

                hrCalcuFreq(bChannel,dwFreq,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new frequency
                m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;   // clear effect
              }
              break;

    // FFx & FEx : fine slide up
    case  6 : {
                BYTE bValue = bInfo;
                if( !bValue)  bValue = m_lpParamCur->m_chanInfo[bChannel].bOldSlide;
                m_lpParamCur->m_chanInfo[bChannel].bOldSlide = bValue;
                if(bValue < 0xe0) break;                              // not a fine slide

                BYTE bLow  = BYTE(bValue & 0xf);
                BYTE bHigh = BYTE(bValue >> 4);
                if(bHigh  != 0xe) bLow <<= 2;                         // fine = 4*extra fine

                DWORD dwFreq = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                dwFreq -= bLow;
                if(dwFreq >= 0x80000000)                              // <0
                   dwFreq  = 0;                                       // cannot go higher than 0
                m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwFreq;

                hrCalcuFreq(bChannel,dwFreq,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new frequency
                m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;   // clear effect
              }
              break;

    // Gxx : port to fix
    case  7 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldPortTo;
                m_lpParamCur->m_chanInfo[bChannel].bOldPortTo = bValue;
              }
              break;

    // Hxy : vibrato fix
    // Uxx : vibrato fix
    case  8 :
    case 21 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVibrato;
                else
                  {
                  // NotOldVib
                  BYTE bLow  = bValue;
                  BYTE bHigh = BYTE(bValue & 0xf0);
                  if( !bHigh)
                    {                                                 // 0: load old x (speed)
                    bHigh  = m_lpParamCur->m_chanInfo[bChannel].bOldVibrato;
                    bHigh &= 0xf0;
                    }
                  bLow  &= 0xf;
                  bValue = BYTE(bHigh | bLow);
                  }
                m_lpParamCur->m_chanInfo[bChannel].bOldVibrato = bValue;
              }
              break;

    // Jxy : arpeggio fix
    case 10 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldArpeggio;
                m_lpParamCur->m_chanInfo[bChannel].bOldArpeggio = bValue;

                BYTE  bNote = m_lpParamCur->m_chanInfo[bChannel].bLastNote;
                DWORD dwArp = m_dwNoteFreq[bNote];                    // 1st arpeggio note
                m_lpParamCur->m_chanInfo[bChannel].dwArpVals[0] = dwArp;

                BYTE  bLow  = BYTE(bValue & 0xf);
                BYTE  bHigh = BYTE(bValue >> 4);

                BYTE  bNext = BYTE(bNote+bHigh);
                BYTE  bMax  = sizeof(m_dwNoteFreq)/sizeof(DWORD);
                if(bNext >= bMax) bNext -= bMax;                      // 120 notes (10*12)
                dwArp = m_dwNoteFreq[bNext];
                m_lpParamCur->m_chanInfo[bChannel].dwArpVals[1] = dwArp;

                bNext = BYTE(bNote+bLow);
                if(bNext >= bMax) bNext -= bMax;
                dwArp = m_dwNoteFreq[bNext];
                m_lpParamCur->m_chanInfo[bChannel].dwArpVals[2] = dwArp;
              }
              break;

    // Kxy : vib vslide fix
    // Lxy : vib vslide fix
    case 11 :
    case 12 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide;
                m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide = bValue;
              }
              break;

    // Oxx : sample offset
    case 15 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldOffset;
                m_lpParamCur->m_chanInfo[bChannel].bOldOffset = bValue;

                DWORD dwOff = bValue << 8;
                if(dwOff >= m_lpParamCur->m_chanInfo[bChannel].dwLen)
                  {
                  if(!m_lpParamCur->m_chanInfo[bChannel].dwRepLen)
                    {                                                 // no loop
                    dwOff = m_lpParamCur->m_chanInfo[bChannel].dwLen-1;
                    }                                                 // end of sample
                  else
                    {                                                 // loop
                    dwOff -= m_lpParamCur->m_chanInfo[bChannel].dwLen;

                    do
                      {
                      if(dwOff < m_lpParamCur->m_chanInfo[bChannel].dwRepLen) break;
                      dwOff -= m_lpParamCur->m_chanInfo[bChannel].dwRepLen;
                      }
                    while(true);

                    dwOff += m_lpParamCur->m_chanInfo[bChannel].dwRepeate;
                    }
                  }
                m_lpParamCur->m_chanInfo[bChannel].dwOffsetAdd = dwOff;
                m_lpParamCur->m_chanInfo[bChannel].dwError     = 0;
              }
              break;

    // Qxy : retrig fix
    case 17 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldRetrig;
                m_lpParamCur->m_chanInfo[bChannel].bOldRetrig = bValue;

                BYTE bLow = BYTE(bValue & 0xf);
                if( !bLow)
                  {                                                   // 0: do nothing
                  m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;
                  break;
                  }

                if(m_lpParamCur->m_chanInfo[bChannel].bOldEffect != 17)
                  {                                                   // reset retrig
                  m_lpParamCur->m_chanInfo[bChannel].bEffectTime = bLow;
                  m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 5; // new note+volume
                  break;
                  }

                m_lpParamCur->m_chanInfo[bChannel].bEffectTime--;
                if(m_lpParamCur->m_chanInfo[bChannel].bEffectTime) break; 
                                                                      // not finished
                m_lpParamCur->m_chanInfo[bChannel].bEffectTime = bLow;
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 5;
                vRetrigSlide(bChannel,bValue);                        // test volume slide
              }
              break;

    // Rxy : tremolo fix
    case 18 : {
                BYTE bValue = bInfo;
                if( !bValue) bValue = m_lpParamCur->m_chanInfo[bChannel].bOldTremolo;
                else
                  {
                  // NotOldTrem
                  BYTE bLow  = bValue;
                  BYTE bHigh = BYTE(bValue & 0xf0);
                  if( !bHigh)
                    {                                                 // 0: load old x (speed)
                    bHigh  = m_lpParamCur->m_chanInfo[bChannel].bOldTremolo;
                    bHigh &= 0xf0;
                    }
                  bLow  &= 0xf;
                  bValue = BYTE(bHigh | bLow);
                  }
                m_lpParamCur->m_chanInfo[bChannel].bOldTremolo = bValue;
              }
              break;

    // Sny : special fix
    case 19 : {
                BYTE  bLow  = BYTE(bInfo & 0xf);
                BYTE  bHigh = BYTE(bInfo >> 4);

                if(bHigh == 8)
                  {                                                   // set pan position (S8x)
                  if(!m_lpParamCur->m_boStereo) break;
                  m_lpParamCur->m_bPanPos[bChannel] = bLow;
                  }

                else if(bHigh == 0xc)
                  {                                                   // cut note (SCx)
                  if(!bLow) 
                    m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;
                  else
                    {
                    if(bLow >= m_lpParamCur->m_bPatternSpeed)         // delay>=speed
                      m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;
                    else                                              // save cut delay
                      m_lpParamCur->m_chanInfo[bChannel].bEffectTime = bLow;
                    }
                  }

                else m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;
              }                                                       // clear effect
              break;

    // Txx : set tempo (21-FF)
    case 20 : if(bInfo < 0x21) break;                                 // min=0x21
              m_lpParamCur->m_bTempo = bInfo;
              //hrSetBPM(bInfo);
              break;

    // Vxx : set global vol (00-40)
    case 22 : if(bInfo > 0x40) break;                                 // max=64
              m_lpParamCur->m_bGlobalVolume = bInfo;
              break;

    // Zxx : synchro demo (NOT an S3M effect !)
    case 26 : m_lpParamCur->m_bSynchro = bInfo;
              break;

    //
    default : break;
    }

  return(_ERR_(_OK_));
  }

/************************************************ hrTickEffects *******************************************************/
// process tick effects
// 13/10/99, M
// in : channel,effect,info
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrTickEffects(const BYTE bChannel,const BYTE bCommand,const BYTE bInfo,const DWORD dwMix)
  {
  _DTRCDEF_("CSoundTrack::hrTickEffects");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  switch(bCommand)
    {
    // Dx0 & D0y : vol slide
    case  4 : {
                BYTE bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide;
                BYTE bLow   = BYTE(bValue & 0xf);
                BYTE bHigh  = BYTE(bValue >> 4);

                if(!bLow)
                  {                                                   // Dx0
                  // VolSlideUp
                  bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bLow += bHigh;
                  if(bLow >= 0x40) bLow = 0x3f;                       // max=63
                  }

                else
                  {
                  // VolSlideDown
                  bHigh  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bHigh -= bLow;
                  if(bHigh >= 0x80) bHigh = 0;                        // min=0
                  bLow = bHigh;
                  }

                m_lpParamCur->m_chanInfo[bChannel].bVol        = bLow;
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;
              }
              break;

    // Exx : slide down
    case  5 : {
                DWORD dwValue = m_lpParamCur->m_chanInfo[bChannel].bOldSlide;
                dwValue <<= 2;
                DWORD dwFreq  = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                dwFreq += dwValue;

                if(dwFreq > m_dwNoteFreq[0])
                   dwFreq = m_dwNoteFreq[0];                          // cannot go below C0 note
                m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwFreq;

                hrCalcuFreq(bChannel,dwFreq,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new frequency
              }
              break;

    // Fxx : slide up
    case  6 : {
                DWORD dwValue = m_lpParamCur->m_chanInfo[bChannel].bOldSlide;
                dwValue <<= 2;
                DWORD dwFreq  = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                dwFreq -= dwValue;

                if(dwFreq >= 0x80000000)                              // <0
                   dwFreq  = 0;                                       // cannot go higher than 0
                m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwFreq;

                hrCalcuFreq(bChannel,dwFreq,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new frequency
              }
              break;

    // Gxx : port to tone
    case  7 : {
                DWORD dwValue  = m_lpParamCur->m_chanInfo[bChannel].bOldPortTo;
                dwValue <<= 2;
                DWORD dwFreq   = m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                DWORD dwPortTo = m_lpParamCur->m_chanInfo[bChannel].dwPortTo;

                if(dwFreq < dwPortTo)
                  {
                  dwFreq += dwValue;                                  // go up
                  if(dwFreq > dwPortTo)                               // ...but not too far
                    m_lpParamCur->m_chanInfo[bChannel].dwST3Freq = dwPortTo;
                  }

                if(dwFreq > dwPortTo)
                  {
                  dwFreq -= dwValue;                                  // go down
                  if(dwFreq >= 0x80000000)                            // ...but not too far
                    dwFreq = dwPortTo;

                  if(dwFreq < dwPortTo)
                    dwFreq = dwPortTo;
                  }

                m_lpParamCur->m_chanInfo[bChannel].dwST3Freq   = dwFreq;
                hrCalcuFreq(bChannel,dwFreq,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new freq
              }
              break;

    // Hxy : vibrato
    // Uxy : vibrato
    case  8 :
    case 21 : {
                BYTE bVibrato = m_lpParamCur->m_chanInfo[bChannel].bOldVibrato;
                BYTE bHigh    = BYTE(bVibrato & 0xf0);
                BYTE bLow     = BYTE(bVibrato & 0xf);                           // depth
                bHigh       >>= 2;                                              // speed*4 (*16/4)
                BYTE bVibPos  = m_lpParamCur->m_chanInfo[bChannel].bVibratoPos;
                bVibPos      += bHigh;
                m_lpParamCur->m_chanInfo[bChannel].bVibratoPos = bVibPos;       // 0..255
                bVibPos     >>= 2;                                              // 0..63
                bVibPos      &= 0x1f;                                           // 32 values

                WORD wSinus   = m_bSinusTable[bVibPos];
                wSinus       *= bLow;
                wSinus      >>= 7;

                if(m_lpParamCur->m_chanInfo[bChannel].bEffectComm != 21)        // not fine vibrato
                  wSinus <<= 2;

                if(m_lpParamCur->m_chanInfo[bChannel].bVibratoPos & 128)        // negative part of sinus
                  wSinus = -wSinus;

                wSinus += (WORD)m_lpParamCur->m_chanInfo[bChannel].dwST3Freq;
                hrCalcuFreq(bChannel,wSinus,dwMix);                             // calculate freq, don't save
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;             // new freq  
              }
              break;

    // Jxy : arpeggio
    case 10 : {
                BYTE  bCount = m_lpParamCur->m_chanInfo[bChannel].bArpCounter;
                DWORD dwArp  = m_lpParamCur->m_chanInfo[bChannel].dwArpVals[bCount];
                DWORD dwFine = m_lpParamCur->m_chanInfo[bChannel].dwFine;

                if(dwFine)
                  {
                  dwArp *= _MIDC_RATE_;
                  dwArp /= dwFine;
                  }
                else dwArp = 0;

                hrCalcuFreq(bChannel,dwArp,dwMix);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new freq
              }
              break;

    // Kxy : vib vslide
    case 11 : {
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;   // new freq
                BYTE bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide;
                BYTE bHigh  = BYTE(bValue >> 4);
                BYTE bLow   = BYTE(bValue & 0xf);

                if(bHigh == 0xf) break;                               // fine volume slide not allowed here
                if(bLow  == 0xf) break;

                if(bLow)
                  {
                  bHigh  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bHigh -= bLow;
                  if(bHigh >= 0x80) bHigh = 0;                        // min=0
                  bLow   = bHigh;
                  }

                else
                  {
                  bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bLow += bHigh;
                  if(bLow >= 0x40) bLow = 0x3f;                       // max=63
                  }

                m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;   // new volume

                hrTickEffects(bChannel,8,bInfo,dwMix);                // vibrato
              }
              break;

    // Lxy : port to slide
    case 12 : {
                BYTE bValue = m_lpParamCur->m_chanInfo[bChannel].bOldVolSlide;
                BYTE bHigh  = BYTE(bValue >> 4);
                BYTE bLow   = BYTE(bValue & 0xf);

                if(bHigh == 0xf) break;                               // fine volume slide not allowed here
                if(bLow  == 0xf) break;

                if(bLow)
                  {
                  bHigh  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bHigh -= bLow;
                  if(bHigh >= 0x80) bHigh = 0;                        // min=0
                  bLow   = bHigh;
                  }

                else
                  {
                  bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
                  bLow += bHigh;
                  if(bLow >= 0x40) bLow = 0x3f;                       // max=63
                  }

                m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;   // new volume

                hrTickEffects(bChannel,7,bInfo,dwMix);                // port to tone
              }
              break;

    // Qxy : retrig
    case 17 : {
                if(m_lpParamCur->m_chanInfo[bChannel].bEffectTime) break;       // not time for a retrig
                BYTE bValue = m_lpParamCur->m_chanInfo[bChannel].bOldRetrig;
                BYTE bLow   = BYTE(bValue & 0xf);

                m_lpParamCur->m_chanInfo[bChannel].bEffectTime = bLow;          // reset retrig delay
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 5;             // new note+volume
                vRetrigSlide(bChannel,bValue);
              }
              break;

    // Rxy : tremolo
    case 18 : {
                BYTE bTremolo = m_lpParamCur->m_chanInfo[bChannel].bOldTremolo;
                BYTE bHigh    = BYTE(bTremolo & 0xf0);
                BYTE bLow     = BYTE(bTremolo & 0xf);
                bHigh       >>=2;                                               // speed*4 (*16/4)
                BYTE bTremPos = m_lpParamCur->m_chanInfo[bChannel].bTremoloPos;
                bTremPos     += bHigh;
                m_lpParamCur->m_chanInfo[bChannel].bTremoloPos = bTremPos;
                bTremPos    >>= 2;
                bTremPos     &= 0x1f;                                           // 32 values

                WORD wSinus   = m_bSinusTable[bTremPos];
                wSinus       *= bLow;
                wSinus      >>= 7;

                if(m_lpParamCur->m_chanInfo[bChannel].bTremoloPos & 128)        // negative part of sinus
                  wSinus = -wSinus;

                m_lpParamCur->m_chanInfo[bChannel].bTremolo    = BYTE(wSinus & 0xff);
                m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 4;             // new volume
              }
              break;

    // Sny : special
    case 19 : if(bInfo != 0xc) break;
              if(m_lpParamCur->m_chanInfo[bChannel].bEffectTime) break;         // not time to cut note

              m_lpParamCur->m_chanInfo[bChannel].dwGusFreq   = 0;               // no sound
              m_lpParamCur->m_chanInfo[bChannel].dwGusRem    = 0;
              m_lpParamCur->m_chanInfo[bChannel].bParam2Set |= 2;               // new freq
              m_lpParamCur->m_chanInfo[bChannel].bEffectComm = 0;               // clear effect
              break;

    //
    default : break;
    }

  return(_ERR_(_OK_));
  }

/************************************************ vRetrigSlide ********************************************************/
// test volume slide
// 13/10/99, M
// in : channel,xy info
// out:
/**********************************************************************************************************************/

void CSoundTrack::vRetrigSlide(const BYTE bChannel,const BYTE bValue)
  {
  _DTRCDEF_("CSoundTrack::vRetrigSlide");
  if(!m_lpParamCur) return;

  BYTE bHigh = BYTE(bValue >> 4);
  BYTE bLow;
  if(!bHigh) return;                                        // x=0

  if(bHigh > 7)
    {
    // RetSlideUp
    bHigh -= 8;
    if(!bHigh) return;                                      // 8: undefined

    if(bHigh > 5)
      {                                                     // x>0xd
      // NoUpSlide
      bHigh -= 6;
      if(!bHigh)
        {
        bLow   = m_lpParamCur->m_chanInfo[bChannel].bVol;
        bLow  *= 3;
        bLow >>= 1;                                         // x=0xe : *3/2
        }

      else
        {
        // RetUpTwice
        bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
        bLow += bLow;                                       // x=0xf : *2
        }

      // RetHighTst
      if(bLow >= 0x40) bLow = 0x3f;
      m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
      return;
      }

    bHigh--;
    bLow  = BYTE(1 << bHigh);                               // x=9..0xd : vol+=2^(x-9)
    bLow += m_lpParamCur->m_chanInfo[bChannel].bVol;
    if(bLow >= 0x40) bLow = 0x3f;
    m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
    return;
    }

  if(bHigh > 5)
    {
    // NoDownSlide
    bHigh -= 6;
    if(!bHigh)
      {
      bLow  = m_lpParamCur->m_chanInfo[bChannel].bVol;
      bLow += bLow;
      bLow /= 3;                                            // x=6 : *2/3
      }

    else
      {
      // NotTwoThird
      bLow   = m_lpParamCur->m_chanInfo[bChannel].bVol;
      bLow >>= 1;                                           // x=7 : *1/2
      }

    m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
    return;
    }

  bHigh--;
  bLow  = BYTE(1 << bHigh);                                 // x=9..0xd : vol+=2^(x-9)
  bLow  = BYTE(m_lpParamCur->m_chanInfo[bChannel].bVol - bLow);
  if(bLow >= 0x80) bLow = 0;
  m_lpParamCur->m_chanInfo[bChannel].bVol = bLow;
  return;
  }

/************************************************ hrCalcuFreq *********************************************************/
// calculate real mixing frequency
// 10/10/99, M
// in : channel,note rate,mix speed
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrCalcuFreq(const BYTE bChannel,const DWORD dwRate,const DWORD dwMix)
  {
  _DTRCDEF_("CSoundTrack::hrCalcuFreq");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  DWORD dwFreq = 0;
  DWORD dwRem  = 0;

  if(dwRate)
    {
    dwFreq   = 128*(DWORD)_MIDC_RATE_*1712*2;
    dwFreq  /= dwRate;

    DWORD dwMixSpeed = dwMix << 8;
    dwRem    = dwFreq % dwMixSpeed;
    dwFreq  /= dwMixSpeed;

    DWORD dwRatio = (DWORD)-1;
    dwRatio /= dwMixSpeed;
    dwRem   *= dwRatio;                                     // dwRem in 0..2^32
    }

  // NoFreq

  m_lpParamCur->m_chanInfo[bChannel].dwGusFreq = dwFreq;
  m_lpParamCur->m_chanInfo[bChannel].dwGusRem  = dwRem;

  return(_ERR_(_OK_));
  }

/************************************************ hrNextOrder *********************************************************/
// jump to next order
// 14/10/99, M
// in :
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrNextOrder()
  {
  _DTRCDEF_("CSoundTrack::hrNextOrder");
  if(!m_lpParamCur) return(_ERR_(_S3MERR_PARAMCUR_));

  m_lpParamCur->m_bPatternCount = 1;
  m_lpParamCur->m_bPatternRow   = 64;
  return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                PARAMS SONG                                                         */
/**********************************************************************************************************************/

/************************************************ hrSetCurParams ******************************************************/
// set m_lpParamCur to nth params set in the list
// 02/12/99, M
// in : set number
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrSetCurParams(const DWORD dwNb)
  {
  _DTRCDEF_("CSoundTrack::hrSetCurParams");
  CParamsSong* lpCur = m_lpParamRoot;
  DWORD        dwNum = dwNb;

  while(lpCur->m_lpNext && dwNum)
    {
    lpCur = lpCur->m_lpNext;
    dwNum--;
    }
  if(dwNum) return(_ERR_(_S3MERR_PARAMCUR_));               // not found, dwNb to big

  m_lpParamCur = lpCur;
  return(_ERR_(_OK_));
  }

/**********************************************************************************************************************/
/*                                                SOUNDS                                                              */
/**********************************************************************************************************************/

/************************************************ hrSetPanning ********************************************************/
// change sound panning
// 03/12/99, M
// in : sound handle,new panning
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrSetPanning(const DWORD dwHandle,const BYTE bPan)
  {
  _DTRCDEF_("CSoundTrack::hrSetPanning");
  if(bPan > 0x0f) return(_ERR_(_S3MERR_BADPARAM_));
  CParamsSong* lpCur   = m_lpParamRoot;
  bool         boFound = false;

  // search handle

  while(lpCur)
    {
    if(lpCur->m_dwHandle == dwHandle)
      {
      boFound = true;
      break;
      }
    lpCur = lpCur->m_lpNext;
    }
  if(!boFound) return(_ERR_(_S3MERR_BADHANDLE_));

  // OK

  for(DWORD dwI = 0; dwI < m_dwChannels; dwI++)
    lpCur->m_bPanPos[dwI] = bPan;

  return(_ERR_(_OK_));
  }

/************************************************ hrSetVolume *********************************************************/
// change sound volume
// 03/12/99, M
// in : sound handle,new volume
// out: _OK_ or error ID
/**********************************************************************************************************************/

HRESULT CSoundTrack::hrSetVolume(const DWORD dwHandle,const BYTE bVol)
  {
  _DTRCDEF_("CSoundTrack::hrSetVolume");
  CParamsSong* lpCur   = m_lpParamRoot;
  bool         boFound = false;

  // search handle

  while(lpCur)
    {
    if(lpCur->m_dwHandle == dwHandle)
      {
      boFound = true;
      break;
      }
    lpCur = lpCur->m_lpNext;
    }
  if(!boFound) return(_ERR_(_S3MERR_BADHANDLE_));

  // OK

  lpCur->m_bGlobalVolume = (bVol > 64) ? BYTE(64) : bVol;
  return(_ERR_(_OK_));
  }
