/**********************************************************************************************************************/
/*                                                SOUNDTRACK CLASS                                                    */
/* 03/10/99, M : creation of the file                                                                                 */
/*               handles a generic soundtrack (effects...)                                                            */
/* 01/12/99, M : params song                                                                                          */
/**********************************************************************************************************************/

#if !defined(AFX_SOUNDTRACK_H__4C0D8F3A_79A8_11D3_9848_FD35403E9742__INCLUDED_)
#define AFX_SOUNDTRACK_H__4C0D8F3A_79A8_11D3_9848_FD35403E9742__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif

/************************************************ INCLUDE *************************************************************/

#include  "ParamsSong.h"

/************************************************ DEFINE **************************************************************/

// values

#define   _MAX_ORDERS_                  255                 // maximum number of orders
#define   _MAX_INSTRUMENTS_             99                  // maximum number of instruments
#define   _MAX_PATTERNS_                100                 // maximum number of patterns

#define   _MIDC_RATE_                   8363                // C4 default speed

/************************************************ TYPEDEF & STRUCT ****************************************************/

/************************************************ CLASSES *************************************************************/

class     CInstrument;

/**********************************************************************************************************************/
/*                                                SoundTrack                                                          */
/**********************************************************************************************************************/

class     CSoundTrack
  {
// public methods

  public:

    // constructors & destructor

	  CSoundTrack                         (void); 
	  virtual        ~CSoundTrack         (void);

    void            vInit               (void);
    void            vClean              (void);

    // load & inits

    HRESULT         hrUnload            (void);                       // free   file memory
    HRESULT         hrLoad              (const char* lpszName);       // load   file from disk
    virtual const char* 
                    lpszGetExtension    (void) const;                 // return file extension
    virtual HRESULT hrCheckSignature    (void);                       // check  file type
    virtual HRESULT hrInitVars          (void);                       // init starting values + samples & patterns pointers

    // play

    HRESULT         hrStartPlaying      (const BYTE  bOrder,  const BYTE bRow,const bool boStereo,const bool  boRepeat=true,
                                         const bool boNew=false,DWORD* lpdwHandle=NULL,CParamsSong** lplpParams=NULL);
    HRESULT         hrStopPlaying       (void);                       // stop
    HRESULT         hrProcessEffects    (const DWORD dwMix);          // effects
    HRESULT         hrPattEffects       (const BYTE  bChannel,const BYTE bCommand,const BYTE  bInfo,const DWORD dwMix);
    HRESULT         hrTickEffects       (const BYTE  bChannel,const BYTE bCommand,const BYTE  bInfo,const DWORD dwMix);
    void            vRetrigSlide        (const BYTE  bChannel,const BYTE bValue);
    HRESULT         hrCalcuFreq         (const BYTE  bChannel,const DWORD dwRate ,const DWORD dwMix);

    HRESULT         hrNextOrder         (void);

    // pattern data

    virtual void    vGoToRow            (BYTE** lplpPattern,const BYTE bRow) = 0;
    virtual void    vUnpackRow          (BYTE*  lpbUnpacked,BYTE** lplpNote) = 0;

    // get

    bool            boGetStereo         (void) const;
    BYTE            bGetMasterVolume    (void) const;
    BYTE            bGetGlobalVolume    (void) const;
    BYTE            bGetSpeed           (void) const;
    BYTE            bGetTempo           (void) const;

    S3M_CHANNEL*    lpGetChannel        (const BYTE bChannel) const;
    BYTE            bGetPanPos          (const BYTE bChannel) const;
    bool            boIsChanActive      (const BYTE bChannel) const;

    BYTE            bGetOrder           (void) const;
    BYTE            bGetNbOrders        (void) const;
    BYTE            bGetRow             (void) const;
    BYTE            bGetSynchro         (void) const;
    bool            boGetRepeat         (void) const;
    bool            boIsPlaying         (void) const;

    DWORD           dwGetRunningCount   (void) const;
    float           fGetRunningTime     (void) const;
    virtual const char* lpszGetTitle    (void) const = 0;
    DWORD           dwGetFileSize       (void) const;
    DWORD           dwGetChannels       (void) const;

    static DWORD    dwGetMaxNbParams    (void);
    static DWORD    dwGetNbParams       (void);

    // set

    void            vSetChanActive      (const BYTE bChannel,const bool boActive);
    void            vSetPlaying         (const bool boPlaying);
    void            vSetGlobalVolume    (const BYTE bVolume);
    bool            boSetOrder          (const BYTE bOrder);

    // params song

    HRESULT         hrSetCurParams      (const DWORD dwNb);           // try to set params dwNb as the current ones (-1=last)

    // sounds

    HRESULT         hrSetPanning        (const DWORD dwHandle,const BYTE bPan);
    HRESULT         hrSetVolume         (const DWORD dwHandle,const BYTE bVol);

// protected members

  protected:

    // errors

    HRESULT         m_hrError;                              // error ID

    // file

    BYTE*           m_lpbFile;                              // file in memory
    DWORD           m_dwFileSize;                           // file size

    // params song

    CParamsSong*    m_lpParamRoot;                          // linked list of playing params
    CParamsSong*    m_lpParamCur;                           // current params set
    CParamsSong*    m_lpParamLast;                          // last element in list

    static DWORD    m_dwNbParams;                           // nb elements in linked list
    static DWORD    m_dwMaxNbParams;                        // max nb elems
    static DWORD    m_dwHandleCount;                        // next sound handle (always goes up)

    // flags

    bool            m_boInitVars;                           // vars initialized ?

    // song initial playing values

    bool            m_boStereo0;                            // stereo OFF/ON
    BYTE            m_bGlobalVolume0;                       // song volume multiplier (0-64)
    BYTE            m_bPatternSpeed0;                       // number of ticks for each row
    BYTE            m_bTempo0;                              // tick time (default=125 BPM)

    BYTE            m_bPanPos0    [_MAX_CHANNELS_];         // panning positions (0..15)

    // song common values

    BYTE            m_bMasterVolume;                        // postprocessing volume
    BYTE            m_bOrders;                              // nb orders (for display only)
    BYTE            m_bOrderList  [_MAX_ORDERS_];           // orders (pattern numbers)

    CInstrument*    m_lpInstrument[_MAX_INSTRUMENTS_];      // instruments addresses
    BYTE*           m_lpbPattern  [_MAX_PATTERNS_];         // patterns data addresses

    DWORD           m_dwChannels;                           // channels used (disabled or not)
    bool            m_boChanActive[_MAX_CHANNELS_];         // is the channel active ?

    //  arrays

    static DWORD    m_dwNoteFreq  [];                       // notes frequencies
    static BYTE     m_bSinusTable [];                       // sinus values (vibrato/tremolo)
  };

/************************************************ INLINE **************************************************************/

//#include  "SoundTrack.inl"

/**********************************************************************************************************************/

#endif // !defined(AFX_SOUNDTRACK_H__4C0D8F3A_79A8_11D3_9848_FD35403E9742__INCLUDED_)
