/**********************************************************************************************************************/
/*                                                SOUNDTRACK CLASS                                                    */
/* 03/10/99, M : creation of the file                                                                                 */
/*               handles a generic soundtrack (effects...)                                                            */
/**********************************************************************************************************************/

#ifndef   _SOUNDTRACK_INL_
#define   _SOUNDTRACK_INL_

/************************************************ INCLUDE *************************************************************/

#include  "SoundTrack.h"

/**********************************************************************************************************************/
/*                                                GET                                                                 */
/**********************************************************************************************************************/

inline bool CSoundTrack::boGetStereo() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_boStereo : m_boStereo0);
  }

inline BYTE CSoundTrack::bGetMasterVolume() const
  {
  return(m_bMasterVolume);
  }

inline BYTE CSoundTrack::bGetGlobalVolume() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_bGlobalVolume : m_bGlobalVolume0);
  }

inline BYTE CSoundTrack::bGetSpeed() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_bPatternSpeed : m_bPatternSpeed0);
  }

inline BYTE CSoundTrack::bGetTempo() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_bTempo : m_bTempo0);
  }

inline S3M_CHANNEL* CSoundTrack::lpGetChannel(const BYTE bChannel) const
  {
  if(!m_lpParamCur) return(NULL);
  if(bChannel >= _MAX_CHANNELS_) return(NULL);
  return(&(m_lpParamCur->m_chanInfo[bChannel]));
  }

inline BYTE CSoundTrack::bGetPanPos(const BYTE bChannel) const
  {
  if(!m_lpParamCur) return(255);
  if(bChannel >= _MAX_CHANNELS_) return(false);
  return(m_lpParamCur->m_bPanPos[bChannel]);
  }

inline bool CSoundTrack::boIsChanActive(const BYTE bChannel) const
  {
  if(bChannel >= _MAX_CHANNELS_) return(false);
  return(m_boChanActive[bChannel]);
  }

inline BYTE CSoundTrack::bGetOrder() const
  {
  return (m_lpParamCur ? m_lpParamCur->m_bPlayingPattern : BYTE(0));
  }

inline BYTE CSoundTrack::bGetNbOrders() const
  {
  return(m_bOrders);
  }

inline BYTE CSoundTrack::bGetRow() const
  {
  return (m_lpParamCur ? m_lpParamCur->m_bPatternRow : BYTE(0));
  }

inline BYTE CSoundTrack::bGetSynchro() const
  {
  return (m_lpParamCur ? m_lpParamCur->m_bSynchro : BYTE(0));
  }

inline bool CSoundTrack::boGetRepeat() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_boRepeat : false);
  }

inline bool CSoundTrack::boIsPlaying() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_boPlaying : false);
  }

inline DWORD CSoundTrack::dwGetRunningCount() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_dwRunningCount : 0);
  }

inline float CSoundTrack::fGetRunningTime() const
  {
  return(m_lpParamCur ? m_lpParamCur->m_fRunningTime : 0.f);
  }

inline DWORD CSoundTrack::dwGetFileSize() const
  {
  return(m_dwFileSize);
  }

inline DWORD CSoundTrack::dwGetChannels() const
  {
  return(m_dwChannels);
  }

inline DWORD CSoundTrack::dwGetMaxNbParams(void)
  {
  return(m_dwMaxNbParams);
  }

inline DWORD CSoundTrack::dwGetNbParams(void)
  {
  return(m_dwNbParams);
  }

/**********************************************************************************************************************/
/*                                                SET                                                                 */
/**********************************************************************************************************************/

inline void CSoundTrack::vSetChanActive(const BYTE bChannel,const bool boActive)
  {
  if(bChannel >= _MAX_CHANNELS_) return;
  m_boChanActive[bChannel] = boActive;
  }

inline void CSoundTrack::vSetPlaying(const bool boPlaying)
  {
  if(!m_lpParamCur) return;
  m_lpParamCur->m_boPlaying = boPlaying;
  }

inline void CSoundTrack::vSetGlobalVolume(const BYTE bVolume)
  {
  if(!m_lpParamCur) return;
  m_lpParamCur->m_bGlobalVolume = bVolume;
  }

inline bool CSoundTrack::boSetOrder(const BYTE bOrder)
  {
  if(!m_lpParamCur)       return(false);
  if(m_bOrders <= bOrder) return(false);
  m_lpParamCur->m_bPlayingPattern = BYTE(bOrder-1);
  m_lpParamCur->m_bPatternCount   = 1;
  m_lpParamCur->m_bPatternRow     = 64;
  memset(m_lpParamCur->m_chanInfo,0,sizeof(m_lpParamCur->m_chanInfo));
  return(true);
  }

/**********************************************************************************************************************/

#endif
